#include <windows.h>

#include "debug.h"

#include "bgjob.h"

// For all members the values are already set when the thread is started. As
// long as the thread runs, the values are considered as read-only, thus there
// is no need for using a critical section to protect access.
// The only member that *could* be changed at thread runtime is hOwner. Well,
// that should better not be done... for many reasons! When there are changing
// owners, the view on whether the job runs or not is compromised and other
// ugly problems might come up. So, just don't do it; always set the job owner
// before starting the job (aka. resuming the thread).
struct BackgroundJob {
	HANDLE hThread;
	HANDLE hCancelEvent;
	HWND hOwner;
	struct CommDev *lpDevice;
	BGJOBPROC lpProc;
	LPVOID lpProcParam;
};

static void SafeDestroyBackgroundJob(struct BackgroundJob **lpJobPtr)
{
	if (*lpJobPtr)
	{
		if ((*lpJobPtr)->hThread)
			CloseHandle((*lpJobPtr)->hThread);
		if ((*lpJobPtr)->hCancelEvent)
			CloseHandle((*lpJobPtr)->hCancelEvent);
	}
}

static void SendFeedback(HWND hOwner, UINT uMsg, WPARAM wParam)
{
	if (hOwner != NULL)
		PostMessage(hOwner, uMsg, wParam, 0);
}

static DWORD WINAPI BackgroundJobProc(LPVOID lpParameter)
{
	struct BackgroundJob *lpJob = lpParameter;

	TRACE("Job at %p started.\n", lpJob);
	SendFeedback(lpJob->hOwner, WM_JOB_STARTED, 0);

	WORD wResult = BGJOB_RESULT_RUNNING;
	BOOL fCanceled = FALSE;
	while (!fCanceled && wResult == BGJOB_RESULT_RUNNING)
	{
		if (WaitForSingleObject(lpJob->hCancelEvent, 0) == WAIT_TIMEOUT)
		{
			wResult = lpJob->lpProc(lpJob->lpDevice, BGJOB_ACTION_EXECUTE, lpJob->lpProcParam);
			SendFeedback(lpJob->hOwner, WM_JOB_PROGRESS, 0);
		}
		else
		{
			TRACE("Canceling job at %p.\n", lpJob);
			wResult = lpJob->lpProc(lpJob->lpDevice, BGJOB_ACTION_CANCEL, lpJob->lpProcParam);
			fCanceled = TRUE;
		}
	}

	TRACE("Job at %p finished %s, result = %i\n", lpJob, fCanceled ? "by cancelation" : "normally", wResult);
	SendFeedback(lpJob->hOwner, WM_JOB_FINISHED, (WPARAM) wResult);

	return 0;
}

struct BackgroundJob *CreateBackgroundJob(struct CommDev *lpDevice, BGJOBPROC lpProc, LPVOID lpProcParam)
{
	BOOL fFailed = FALSE;

	struct BackgroundJob *lpResult = MALLOC(sizeof(*lpResult));
	fFailed = lpResult == NULL;

	if (!fFailed)
	{
		lpResult->hThread = NULL;
		lpResult->hCancelEvent = NULL;
		lpResult->hOwner = NULL;
		lpResult->lpDevice = lpDevice;
		lpResult->lpProc = lpProc;
		lpResult->lpProcParam = lpProcParam;
	}

	if (!fFailed)
	{
		lpResult->hCancelEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
		fFailed = lpResult->hCancelEvent == NULL;
	}

	if (!fFailed)
	{
		lpResult->hThread = CreateThread(NULL, 0, BackgroundJobProc, lpResult, CREATE_SUSPENDED, NULL);
		fFailed = lpResult->hThread == NULL;
	}

	if (fFailed)
		SafeDestroyBackgroundJob(&lpResult);

	TRACE("Background job created at %p, Thread ID = 0x%x.\n", lpResult, lpResult ? GetThreadId(lpResult->hThread) : 0);
	return lpResult;
}

BOOL SetBackgroundJobOwner(struct BackgroundJob *lpJob, HWND hOwner)
{
	ASSERT(lpJob);
	TRACE("SetBackgroundJobOwner called with lpJob = %p and hOwner = %p.\n", lpJob, hOwner);

	BOOL fResult = FALSE;
	if (lpJob)
	{
		lpJob->hOwner = hOwner;
	}

	return fResult;
}

BOOL StartBackgroundJob(struct BackgroundJob *lpJob)
{
	ASSERT(lpJob);
	TRACE("StartBackgroundJob called with lpJob = %p.\n", lpJob);

	BOOL fResult = FALSE;
	if (lpJob)
	{
		fResult = ResumeThread(lpJob->hThread) != (DWORD) -1;
	}

	return fResult;
}

BOOL CancelBackgroundJob(struct BackgroundJob *lpJob)
{
	ASSERT(lpJob);
	TRACE("CancelBackgroundJob called with lpJob = %p.\n", lpJob);

	BOOL fResult = FALSE;
	if (lpJob)
	{
		fResult = SetEvent(lpJob->hCancelEvent);
	}

	return fResult;
}

BOOL WaitForBackgroundJob(struct BackgroundJob *lpJob, DWORD dwTimeout)
{
	ASSERT(lpJob);
	TRACE("WaitForBackgroundJob called with lpJob = %p.\n", lpJob);

	BOOL fResult = FALSE;
	if (lpJob)
	{
		WaitForSingleObject(lpJob->hThread, dwTimeout);
	}

	return fResult;
}

void DestroyBackgroundJob(struct BackgroundJob *lpJob)
{
	ASSERT(lpJob);
	TRACE("DestroyBackgroundJob called with lpJob = %p.\n", lpJob);

	if (lpJob)
	{
		SetEvent(lpJob->hCancelEvent);
		WaitForSingleObject(lpJob->hThread, INFINITE);
		SafeDestroyBackgroundJob(&lpJob);
	}
}
