#include <stdio.h>
#include <windows.h>
#include <commctrl.h>

#include "commproto.h"
#include "bgjob.h"
#include "protodef.h"
#include "resource.h"
#include "guiutils.h"
#include "transferdlg.h"

#include "debug.h"

#include "downloadtoolbox.h"

struct DownloadToolboxProgress {
	CRITICAL_SECTION csLock;
	WORD wCurrStep;
	WORD wTotalSteps;
};

#define DEF_BUFFER_SIZE      100
#define DOWNLOAD_BLOCK_SIZE  32

static void TransferDlg_UpdateProgress(HWND hDlg, struct TransferDlgData *lpData)
{
	// Keep lock as short as possible
	struct DownloadToolboxProgress *lpProgressData = lpData->lpProgressData;
	EnterCriticalSection(&lpProgressData->csLock);
	WORD wCurrStep = lpProgressData->wCurrStep;
	WORD wTotalSteps = lpProgressData->wTotalSteps;
	LeaveCriticalSection(&lpProgressData->csLock);

	SendMessage(lpData->hProgressBar, PBM_SETRANGE, 0, MAKELPARAM(0, wTotalSteps));
	SendMessage(lpData->hProgressBar, PBM_SETPOS, wCurrStep, 0);

	char szBuffer[DEF_BUFFER_SIZE];
	int iPercent = (int) (wTotalSteps > 0 ? (wCurrStep * 100.0 / wTotalSteps + 0.5) : 0);
	sprintf_s(szBuffer, sizeof(szBuffer), "%i byte%s / %i byte%s (%i%%)",
		wCurrStep, (wCurrStep == 1 ? "" : "s"),
		wTotalSteps, (wTotalSteps == 1 ? "" : "s"), iPercent);
	SetWindowText(GetDlgItem(hDlg, IDC_STATUS), szBuffer);
}

struct DownloadToolboxJobProcParam {
	HANDLE hFile;
	HANDLE hFileMapping;
	LPBYTE lpImageBuffer;
	WORD wImageSize;
	WORD wCurrPos;
	struct DownloadToolboxProgress *lpProgressData;
};

static WORD DownloadToolboxJobProc(struct CommDev *lpDevice, BYTE bAction, LPVOID lpParam)
{
	ASSERT(lpDevice);
	ASSERT(lpParam);

	WORD wResult = BGJOB_RESULT_FAILURE;

	if (bAction == BGJOB_ACTION_EXECUTE)
	{
		struct DownloadToolboxJobProcParam *lpData = lpParam;

		// Send data in blocks to limit the number of progress indicator updates
		const WORD wNextStop = min(lpData->wCurrPos + DOWNLOAD_BLOCK_SIZE, lpData->wImageSize);
		BOOL fFailed = FALSE;
		while (!fFailed && lpData->wCurrPos < wNextStop)
		{
			if (CommDevWriteByte(lpDevice, lpData->lpImageBuffer[lpData->wCurrPos]))
			{
				lpData->wCurrPos++;
			}
			else
			{
				TRACE("Error during toolbox download at byte %u of %u.\n", lpData->wCurrPos, lpData->wImageSize);
				fFailed = TRUE;
			}
		}

		EnterCriticalSection(&lpData->lpProgressData->csLock);
		lpData->lpProgressData->wCurrStep = lpData->wCurrPos;
		lpData->lpProgressData->wTotalSteps = lpData->wImageSize;
		LeaveCriticalSection(&lpData->lpProgressData->csLock);

		if (fFailed)
		{
			wResult = BGJOB_RESULT_FAILURE;
		}
		else
		{
			wResult = lpData->wCurrPos < lpData->wImageSize ? BGJOB_RESULT_RUNNING : BGJOB_RESULT_SUCCESS;
		}
	}

	return wResult;
}

static void SafeDestroyDownloadToolboxJobProcParam(struct DownloadToolboxJobProcParam **lpParamPtr)
{
	if (*lpParamPtr)
	{
		if ((*lpParamPtr)->lpImageBuffer)
			UnmapViewOfFile((*lpParamPtr)->lpImageBuffer);
		if ((*lpParamPtr)->hFileMapping)
			CloseHandle((*lpParamPtr)->hFileMapping);
		if ((*lpParamPtr)->hFile != INVALID_HANDLE_VALUE)
			CloseHandle((*lpParamPtr)->hFile);
		FREE(*lpParamPtr);
		*lpParamPtr = NULL;
	}
}

static struct DownloadToolboxJobProcParam *CreateDownloadToolboxJobProcParam(HWND hWnd, LPCSTR lpPath, struct DownloadToolboxProgress *lpProgressData)
{
	BOOL fFailed = FALSE;

	struct DownloadToolboxJobProcParam *lpResult = MALLOC(sizeof(*lpResult));
	if (lpResult == NULL)
	{
		FormatMessageBox(hWnd, MB_ICONERROR, "Could not allocated memory for background job parameter.");
		fFailed = TRUE;
	}

	if (!fFailed)
	{
		lpResult->hFile = INVALID_HANDLE_VALUE;
		lpResult->hFileMapping = NULL;
		lpResult->lpImageBuffer = NULL;
		lpResult->wImageSize = 0;
		lpResult->wCurrPos = 0;
		lpResult->lpProgressData = lpProgressData;
	}

	if (!fFailed)
	{
		lpResult->hFile = CreateFile(lpPath, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
		if (lpResult->hFile == INVALID_HANDLE_VALUE)
		{
			FormatMessageBox(hWnd, MB_ICONERROR, "Could not open file: %s", GetLastErrorText());
			fFailed = TRUE;
		}
	}

	DWORD dwSize;
	if (!fFailed)
	{
		dwSize = GetFileSize(lpResult->hFile, NULL);
		if (dwSize == INVALID_FILE_SIZE)
		{
			FormatMessageBox(hWnd, MB_ICONERROR, "Could not get file size: %s", GetLastErrorText());
			fFailed = TRUE;
		}
		else if (dwSize > MAX_TOOLBOX_SIZE)
		{
			FormatMessageBox(hWnd, MB_ICONERROR, "File size exceeds supported maximum of %i bytes.", MAX_TOOLBOX_SIZE);
			fFailed = TRUE;
		}
		else
		{
			lpResult->wImageSize = (WORD) dwSize;
			lpProgressData->wTotalSteps = (WORD) dwSize;
		}
	}

	if (!fFailed)
	{
		lpResult->hFileMapping = CreateFileMapping(lpResult->hFile, NULL, PAGE_READONLY, 0, dwSize, NULL);
		TRACE("Created file mapping with handle 0x%p\n", lpResult->hFileMapping, dwSize);
		fFailed = lpResult->hFileMapping == NULL;
	}

	if (!fFailed)
	{
		lpResult->lpImageBuffer = MapViewOfFile(lpResult->hFileMapping, FILE_MAP_READ, 0, 0, dwSize);
		TRACE("Mapped view of file to 0x%p\n", lpResult->lpImageBuffer);
		fFailed = lpResult->lpImageBuffer == NULL;
	}

	if (fFailed)
		SafeDestroyDownloadToolboxJobProcParam(&lpResult);

	return lpResult;
}

static void DestroyDownloadToolboxJobProcParam(struct DownloadToolboxJobProcParam *lpParam)
{
	ASSERT(lpParam);
	SafeDestroyDownloadToolboxJobProcParam(&lpParam);
}

BOOL DownloadToolBox(HWND hWnd, struct CommDev *lpDevice, LPCSTR lpPath)
{
	BOOL fFailed = FALSE;

	struct DownloadToolboxProgress Progress = { 0 };
	InitializeCriticalSection(&Progress.csLock);

	struct TransferDlgData DlgData = { 0 };
	DlgData.lpSource = lpPath;
	DlgData.lpDestination = "Target (RAM)";
	DlgData.lpProgressData = &Progress;
	DlgData.lpUpdateProgressProc = TransferDlg_UpdateProgress;

	struct DownloadToolboxJobProcParam *lpParam = CreateDownloadToolboxJobProcParam(hWnd, lpPath, &Progress);
	if (lpParam)
	{
		struct BackgroundJob *lpJob = CreateBackgroundJob(lpDevice, DownloadToolboxJobProc, lpParam);
		if (lpJob)
		{
			DlgData.lpJob = lpJob;

			HINSTANCE hInst = (HINSTANCE) GetWindowLongPtr(hWnd, GWLP_HINSTANCE);
			LPCSTR lpTemplate = MAKEINTRESOURCE(IDD_TRANSFERDLG);
			if (DialogBoxParam(hInst, lpTemplate, hWnd, TransferDlgProc, (LPARAM) &DlgData) == BGJOB_RESULT_SUCCESS)
			{
				// Nothing to do
			}
			else
			{
				TRACE("Job ended unhappy\n");
				fFailed = TRUE;
			}

			DestroyBackgroundJob(lpJob);
		}
		else
		{
			TRACE("Could not create background job\n");
			fFailed = TRUE;
		}

		DestroyDownloadToolboxJobProcParam(lpParam);
	}
	else
	{
		fFailed = TRUE;
	}

	DeleteCriticalSection(&Progress.csLock);
	return !fFailed;
}
