#include <stdio.h>
#include <windows.h>

#include "debug.h"

#include "settings.h"

#define DEF_BAUD_RATE                      CBR_9600
#define DEF_DATA_BITS                      8
#define DEF_PARITY                         NOPARITY
#define DEF_STOP_BITS                      ONESTOPBIT
#define DEF_DELAY_BEFORE_WRITE             0
#define DEF_DELAY_AFTER_WRITE              0
#define DEF_DELAY_BEFORE_READ              0
#define DEF_DELAY_AFTER_READ               0
#define DEF_DELAY_BEFORE_READ_AFTER_WRITE  0
#define DEF_DELAY_BEFORE_WRITE_AFTER_READ  0
#define DEF_READ_TIMEOUT                   3000
#define DEF_OUTX_CTS_FLOW                  0
#define DEF_OUTX_DSR_FLOW                  0
#define DEF_DTR_CONTROL                    DTR_CONTROL_ENABLE
#define DEF_RTS_CONTROL                    RTS_CONTROL_ENABLE

#define VENDOR_NAME                        "Zotteljedi"
#define APP_NAME                           "toolbox-client"

static BOOL GetRegistryStringValue(HKEY hParent, LPCSTR lpSubKey, LPCSTR lpName, LPSTR lpValue, DWORD dwSize)
{
	return RegGetValue(hParent, lpSubKey, lpName, RRF_RT_REG_DWORD, NULL, lpValue, &dwSize) == ERROR_SUCCESS;
}

static BOOL GetRegistryDwordValue(HKEY hParent, LPCSTR lpSubKey, LPCSTR lpName, DWORD *lpValue)
{
	DWORD dwSize = sizeof(*lpValue);
	return RegGetValue(hParent, lpSubKey, lpName, RRF_RT_REG_DWORD, NULL, lpValue, &dwSize) == ERROR_SUCCESS;
}

static void ValidatePortName(LPSTR lpValue, DWORD dwSize)
{
	TRACE("ValidatePortName: in %s,", lpValue);

	if (dwSize > 0)
	{
		const struct OptionList *list = GetPortNames();

		BOOL fValid = FALSE;
		for (int i = 0; !fValid && i < list->iCount; i++)
		{
			fValid = strcmp(lpValue, list->lpOptions[i].szLabel) == 0;
		}

		if (!fValid)
		{
			if (list->iCount > 0)
				strcpy_s(lpValue, dwSize, list->lpOptions[0].szLabel);
			else
				lpValue[0] = '\0';
		}
	}

	TRACE(" out %s\n", lpValue);
}

void LoadSettings(struct Settings *lpSettings)
{
	lpSettings->szPortName[0] = '\0';
	lpSettings->dwBaudRate = DEF_BAUD_RATE;
	lpSettings->bDataBits = DEF_DATA_BITS;
	lpSettings->bParity = DEF_PARITY;
	lpSettings->bStopBits = DEF_STOP_BITS;
	LoadTweaksDefaults(lpSettings);

	HKEY hSoftware;
	if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software", 0, KEY_QUERY_VALUE, &hSoftware) == ERROR_SUCCESS)
	{
		HKEY hVendor;
		if (RegOpenKeyEx(hSoftware, VENDOR_NAME, 0, KEY_QUERY_VALUE, &hVendor) == ERROR_SUCCESS)
		{
			DWORD dwValue;

			GetRegistryStringValue(hVendor, APP_NAME, "PortName", lpSettings->szPortName,
				sizeof(lpSettings->szPortName));
			if (GetRegistryDwordValue(hVendor, APP_NAME, "BaudRate", &dwValue))
				lpSettings->dwBaudRate = dwValue;
			if (GetRegistryDwordValue(hVendor, APP_NAME, "DataBits", &dwValue))
				lpSettings->bDataBits = (BYTE) dwValue;
			if (GetRegistryDwordValue(hVendor, APP_NAME, "Parity", &dwValue))
				lpSettings->bParity = (BYTE) dwValue;
			if (GetRegistryDwordValue(hVendor, APP_NAME, "StopBits", &dwValue))
				lpSettings->bStopBits = (BYTE) dwValue;
			if (GetRegistryDwordValue(hVendor, APP_NAME, "DelayBeforeWrite", &dwValue))
				lpSettings->wDelayBeforeWrite = (WORD) dwValue;
			if (GetRegistryDwordValue(hVendor, APP_NAME, "DelayAfterWrite", &dwValue))
				lpSettings->wDelayAfterWrite = (WORD) dwValue;
			if (GetRegistryDwordValue(hVendor, APP_NAME, "DelayBeforeRead", &dwValue))
				lpSettings->wDelayBeforeRead = (WORD) dwValue;
			if (GetRegistryDwordValue(hVendor, APP_NAME, "DelayAfterRead", &dwValue))
				lpSettings->wDelayAfterRead = (WORD) dwValue;
			if (GetRegistryDwordValue(hVendor, APP_NAME, "DelayBeforeReadAfterWrite", &dwValue))
				lpSettings->wDelayBeforeReadAfterWrite = (WORD) dwValue;
			if (GetRegistryDwordValue(hVendor, APP_NAME, "DelayBeforeWriteAfterRead", &dwValue))
				lpSettings->wDelayBeforeWriteAfterRead = (WORD) dwValue;
			if (GetRegistryDwordValue(hVendor, APP_NAME, "ReadTimeout", &dwValue))
				lpSettings->wReadTimeout = (WORD) dwValue;
			if (GetRegistryDwordValue(hVendor, APP_NAME, "OutxCtsFlow", &dwValue))
				lpSettings->fOutxCtsFlow = dwValue ? TRUE : FALSE;
			if (GetRegistryDwordValue(hVendor, APP_NAME, "OutxDsrFlow", &dwValue))
				lpSettings->fOutxDsrFlow = dwValue ? TRUE : FALSE;
			if (GetRegistryDwordValue(hVendor, APP_NAME, "DtrControl", &dwValue))
				lpSettings->bDtrControl = (BYTE) dwValue;
			if (GetRegistryDwordValue(hVendor, APP_NAME, "RtsControl", &dwValue))
				lpSettings->bRtsControl = (BYTE) dwValue;
		}
	}

	ValidatePortName(lpSettings->szPortName, sizeof(lpSettings->szPortName));

	TRACE("LoadSettings: %s, %i/%i/%i/%i from CurrentUser/Software/%s/%s\n",
		lpSettings->szPortName, lpSettings->dwBaudRate,
		lpSettings->bDataBits, lpSettings->bParity, lpSettings->bStopBits,
		VENDOR_NAME, APP_NAME);
}

static void SetRegistryStringValue(HKEY hKey, LPCSTR lpName, LPCSTR lpValue)
{
	// MSDN: If the data is of type REG_SZ, REG_EXPAND_SZ, or REG_MULTI_SZ, cbData
	//       must include the size of the terminating null character or characters.
	RegSetValueEx(hKey, lpName, 0, REG_SZ, (BYTE*) lpValue, (DWORD) (strlen(lpValue) + 1));
}

static void SetRegistryDwordValue(HKEY hKey, LPCSTR lpName, DWORD dwValue)
{
	RegSetValueEx(hKey, lpName, 0, REG_DWORD, (BYTE*) &dwValue, sizeof(dwValue));
}

void SaveSettings(const struct Settings *lpSettings)
{
	TRACE("SaveSettings: %s, %i/%i/%i/%i to CurrentUser/Software/%s/%s\n",
		lpSettings->szPortName, lpSettings->dwBaudRate,
		lpSettings->bDataBits, lpSettings->bParity, lpSettings->bStopBits,
		VENDOR_NAME, APP_NAME);

	HKEY hSoftware = NULL;
	if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software", 0, KEY_WRITE, &hSoftware) != ERROR_SUCCESS)
	{
		TRACE("Error opening key Software\n");
		hSoftware = NULL;
	}

	HKEY hVendor = NULL;
	if (hSoftware)
	{
		if (RegCreateKeyEx(hSoftware, VENDOR_NAME, 0, NULL, 0, KEY_WRITE, NULL, &hVendor, NULL) != ERROR_SUCCESS)
		{
			TRACE("Error creating key %s\n", VENDOR_NAME);
			hVendor = NULL;
		}
	}

	HKEY hApplication = NULL;
	if (hVendor)
	{
		if (RegCreateKeyEx(hVendor, APP_NAME, 0, NULL, 0, KEY_WRITE, NULL, &hApplication, NULL) != ERROR_SUCCESS)
		{
			TRACE("Error creating key %s\n", APP_NAME);
			hApplication = NULL;
		}
	}

	if (hApplication)
	{
		SetRegistryStringValue(hApplication, "PortName", lpSettings->szPortName);
		SetRegistryDwordValue(hApplication, "BaudRate", lpSettings->dwBaudRate);
		SetRegistryDwordValue(hApplication, "DataBits", lpSettings->bDataBits);
		SetRegistryDwordValue(hApplication, "Parity", lpSettings->bParity);
		SetRegistryDwordValue(hApplication, "StopBits", lpSettings->bStopBits);
		SetRegistryDwordValue(hApplication, "DelayBeforeWrite", lpSettings->wDelayBeforeWrite);
		SetRegistryDwordValue(hApplication, "DelayAfterWrite", lpSettings->wDelayAfterWrite);
		SetRegistryDwordValue(hApplication, "DelayBeforeRead", lpSettings->wDelayBeforeRead);
		SetRegistryDwordValue(hApplication, "DelayAfterRead", lpSettings->wDelayAfterRead);
		SetRegistryDwordValue(hApplication, "DelayBeforeReadAfterWrite", lpSettings->wDelayBeforeReadAfterWrite);
		SetRegistryDwordValue(hApplication, "DelayBeforeWriteAfterRead", lpSettings->wDelayBeforeWriteAfterRead);
		SetRegistryDwordValue(hApplication, "ReadTimeout", lpSettings->wReadTimeout);
		SetRegistryDwordValue(hApplication, "OutxCtsFlow", lpSettings->fOutxCtsFlow ? 1 : 0);
		SetRegistryDwordValue(hApplication, "OutxDsrFlow", lpSettings->fOutxDsrFlow ? 1 : 0);
		SetRegistryDwordValue(hApplication, "DtrControl", lpSettings->bDtrControl);
		SetRegistryDwordValue(hApplication, "RtsControl", lpSettings->bRtsControl);
	}

	if (hApplication)
		RegCloseKey(hApplication);
	if (hVendor)
		RegCloseKey(hVendor);
	if (hSoftware)
		RegCloseKey(hSoftware);
}

void LoadTweaksDefaults(struct Settings *lpSettings)
{
	lpSettings->wDelayBeforeWrite = DEF_DELAY_BEFORE_WRITE;
	lpSettings->wDelayAfterWrite = DEF_DELAY_AFTER_WRITE;
	lpSettings->wDelayBeforeRead = DEF_DELAY_BEFORE_READ;
	lpSettings->wDelayAfterRead = DEF_DELAY_AFTER_READ;
	lpSettings->wDelayBeforeReadAfterWrite = DEF_DELAY_BEFORE_READ_AFTER_WRITE;
	lpSettings->wDelayBeforeWriteAfterRead = DEF_DELAY_BEFORE_WRITE_AFTER_READ;
	lpSettings->wReadTimeout = DEF_READ_TIMEOUT;
	lpSettings->fOutxCtsFlow = DEF_OUTX_CTS_FLOW;
	lpSettings->fOutxDsrFlow = DEF_OUTX_DSR_FLOW;
	lpSettings->bDtrControl = DEF_DTR_CONTROL;
	lpSettings->bRtsControl = DEF_RTS_CONTROL;
}

#define MAX_PORT_NAMES 256
static struct Option PortNameOptions[MAX_PORT_NAMES];
const struct OptionList *GetPortNames(void)
{
	int iEntry = 0;

	for (int iPort = 1; iPort <= MAX_PORT_NAMES; iPort++)
	{
		char devname[MAX_PATH];
		sprintf_s(PortNameOptions[iEntry].szLabel, sizeof(PortNameOptions[iEntry].szLabel), "COM%i", iPort);
		if (QueryDosDevice(PortNameOptions[iEntry].szLabel, devname, sizeof(devname)) > 0)
		{
			PortNameOptions[iEntry].dwValue = iPort;
			iEntry++;
		}
	}

	static struct OptionList list = { PortNameOptions };
	list.iCount = iEntry;
	return &list;
}

static const struct Option BaudRateOptions[] = {
	{ "110",  CBR_110  },
	{ "300",  CBR_300  },
	{ "600",  CBR_600  },
	{ "1200", CBR_1200 },
	{ "2400", CBR_2400 },
	{ "4800", CBR_4800 },
	{ "9600", CBR_9600 },
};
const struct OptionList *GetBaudRateSettings(void)
{
	static const struct OptionList list = {
		BaudRateOptions,
		ARRAYSIZE(BaudRateOptions)
	};
	return &list;
}

static const struct Option DataBitsOptions[] = {
	{ "7",  7 },
	{ "8",  8 },
};
const struct OptionList *GetDataBitsSettings(void)
{
	static const struct OptionList list = {
		DataBitsOptions,
		ARRAYSIZE(DataBitsOptions)
	};
	return &list;
}

static const struct Option ParityOptions[] = {
	{ "None",  NOPARITY   },
	{ "Odd",   ODDPARITY  },
	{ "Even",  EVENPARITY },
};
const struct OptionList *GetParitySettings(void)
{
	static const struct OptionList list = {
		ParityOptions,
		ARRAYSIZE(ParityOptions)
	};
	return &list;
}

static const struct Option StopBitsOptions[] = {
	{ "1",  ONESTOPBIT  },
	{ "2",  TWOSTOPBITS },
};
const struct OptionList *GetStopBitsSettings(void)
{
	static const struct OptionList list = {
		StopBitsOptions,
		ARRAYSIZE(StopBitsOptions)
	};
	return &list;
}

static const struct Option DtrControlOptions[] = {
	{ "Disable",    DTR_CONTROL_DISABLE   },
	{ "Enable",     DTR_CONTROL_ENABLE    },
	{ "Handshake",  DTR_CONTROL_HANDSHAKE },
};
const struct OptionList *GetDtrControlSettings(void)
{
	static const struct OptionList list = {
		DtrControlOptions,
		ARRAYSIZE(DtrControlOptions)
	};
	return &list;
}

static const struct Option RtsControlOptions[] = {
	{ "Disable",    RTS_CONTROL_DISABLE   },
	{ "Enable",     RTS_CONTROL_ENABLE    },
	{ "Handshake",  RTS_CONTROL_HANDSHAKE },
	{ "Toggle",     RTS_CONTROL_TOGGLE    },
};
const struct OptionList *GetRtsControlSettings(void)
{
	static const struct OptionList list = {
		RtsControlOptions,
		ARRAYSIZE(RtsControlOptions)
	};
	return &list;
}
