#include <stdio.h>
#include <windows.h>
#include <commctrl.h>

#include "commproto.h"
#include "bgjob.h"
#include "protodef.h"
#include "resource.h"
#include "guiutils.h"
#include "transferdlg.h"

#include "debug.h"

#include "transferfile.h"


///////////////////////////////////////////////////////////////////////////////////////////////////
// Generic part for both directions (recv and send)

struct TransferFileProgress {
	CRITICAL_SECTION csLock;
	DWORD dwBytesProcessed;
	DWORD dwTotalBytes;
};

#define DEF_BUFFER_SIZE 100

static void TransferDlg_UpdateProgress(HWND hDlg, struct TransferDlgData *lpData)
{
	// Keep lock as short as possible
	struct TransferFileProgress *lpProgressData = lpData->lpProgressData;
	EnterCriticalSection(&lpProgressData->csLock);
	DWORD dwBytesProcessed = lpProgressData->dwBytesProcessed;
	DWORD dwTotalBytes = lpProgressData->dwTotalBytes;
	LeaveCriticalSection(&lpProgressData->csLock);

	char szBuffer[DEF_BUFFER_SIZE];
	int iPercent = (int) (dwTotalBytes > 0 ? (dwBytesProcessed * 100.0 / dwTotalBytes + 0.5) : 0);

	if (dwBytesProcessed > 0)
	{
		DWORD dwTicksPassed = GetTickCount() - lpData->dwBeginOfTransfer;
		double dTicksPerStep = dwTicksPassed / (double) dwBytesProcessed;
		DWORD dwTicksLeft = (DWORD) (dTicksPerStep * dwTotalBytes) - dwTicksPassed;
		int iSecondsLeft = (dwTicksLeft + 750) / 1000; // Staying too long at "00:00" looks stupid

		sprintf_s(szBuffer, sizeof(szBuffer), "%lu byte%s / %lu byte%s (%i%%), %i:%.2i left",
			dwBytesProcessed, (dwBytesProcessed == 1 ? "" : "s"),
			dwTotalBytes, (dwTotalBytes == 1 ? "" : "s"),
			iPercent, iSecondsLeft / 60, iSecondsLeft % 60);
	}
	else
	{
		sprintf_s(szBuffer, sizeof(szBuffer), "%lu byte%s / %lu byte%s (%i%%)",
			dwBytesProcessed, (dwBytesProcessed == 1 ? "" : "s"),
			dwTotalBytes, (dwTotalBytes == 1 ? "" : "s"), iPercent);
	}

	SetWindowText(GetDlgItem(hDlg, IDC_STATUS), szBuffer);

	// 1000 steps should result in a reasonable smooth progress bar
	double dScaleFactor = dwTotalBytes > 0 ? 1000.0 / dwTotalBytes : 0;
	SendMessage(lpData->hProgressBar, PBM_SETRANGE, 0, MAKELPARAM(0, 1000));
	SendMessage(lpData->hProgressBar, PBM_SETPOS, (WPARAM) (dwBytesProcessed * dScaleFactor + 0.5), 0);
}

struct TransferFileJobProcParam {
	HANDLE hLocalFile;
	LPCSTR lpRemotePath;
	BOOL fRemoteFileOpen;
	WORD wRemoteFileHandle;
	BOOL fSourceSizeKnown;
	DWORD dwBytesProcessed;
	DWORD dwTotalBytes;
	struct TransferFileProgress *lpProgressData;
};

static void SafeDestroyTransferFileJobProcParam(struct TransferFileJobProcParam **lpParamPtr)
{
	if (*lpParamPtr)
	{
		if ((*lpParamPtr)->hLocalFile != INVALID_HANDLE_VALUE)
			CloseHandle((*lpParamPtr)->hLocalFile);
		FREE(*lpParamPtr);
		*lpParamPtr = NULL;
	}
}

static void DestroyTransferFileJobProcParam(struct TransferFileJobProcParam *lpParam)
{
	ASSERT(lpParam);
	SafeDestroyTransferFileJobProcParam(&lpParam);
}

static BOOL OpenRemoteFile(struct CommDev *lpDevice, LPCSTR lpPath, BYTE bAccessMode, LPWORD lpHandle)
{
	BOOL fResult = FALSE;

	struct OpenFileRequest request = { CMD_OPEN_FILE };
	request.bAccessMode = bAccessMode;
	strcpy_s((char*) request.szPath, sizeof(request.szPath), lpPath);
	// bRequestCode + bAccessMode + string length + '\0'
	const WORD wRequestSize = (WORD) (3 + strlen((const char*) request.szPath));

	struct OpenFileResponse response;

	WORD wResponseSize = sizeof(response);
	if (ProtocolTransaction(lpDevice, &request, wRequestSize, &response, &wResponseSize))
	{
		if (response.bDosStatusCode == DOS_CODE_SUCCESS)
		{
			TRACE("Remote file handle = 0x%.4x\n", response.wFileHandle);
			*lpHandle = response.wFileHandle;
			fResult = TRUE;
		}
		else
		{
			TRACE("Negative DOS status code: 0x%.2u\n", response.bDosStatusCode);
		}
	}

	return fResult;
}

static BOOL CloseRemoteFile(struct CommDev *lpDevice, WORD wHandle)
{
	BOOL fResult = FALSE;

	struct CloseFileRequest request = { CMD_CLOSE_FILE };
	request.wFileHandle = wHandle;

	struct CloseFileResponse response;

	WORD wResponseSize = sizeof(response);
	if (ProtocolTransaction(lpDevice, &request, sizeof(request), &response, &wResponseSize))
	{
		if (response.bDosStatusCode == DOS_CODE_SUCCESS)
			fResult = TRUE;
		else
			TRACE("Negative DOS status code: 0x%.2u\n", response.bDosStatusCode);
	}

	return fResult;
}

static BOOL GetRemoteFileSize(struct CommDev *lpDevice, WORD wHandle, LPDWORD lpSize)
{
	BOOL fResult = FALSE;

	struct GetFileSizeRequest request = { CMD_GET_FILE_SIZE };
	request.wFileHandle = wHandle;

	struct GetFileSizeResponse response;

	WORD wResponseSize = sizeof(response);
	if (ProtocolTransaction(lpDevice, &request, sizeof(request), &response, &wResponseSize))
	{
		if (response.bDosStatusCode == DOS_CODE_SUCCESS)
		{
			TRACE("Got remote file size = %lu\n", response.dwFileSize);
			*lpSize = response.dwFileSize;
			fResult = TRUE;
		}
		else
		{
			TRACE("Negative DOS status code: 0x%.2u\n", response.bDosStatusCode);
		}
	}

	return fResult;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Specific part for receiving file from target

static struct TransferFileJobProcParam *CreateRecvFileJobProcParam(HWND hWnd, LPCSTR lpSourcePath,
	LPCSTR lpDestinationPath, struct TransferFileProgress *lpProgressData)
{
	BOOL fFailed = FALSE;

	struct TransferFileJobProcParam *lpResult = MALLOC(sizeof(*lpResult));
	if (lpResult == NULL)
	{
		FormatMessageBox(hWnd, MB_ICONERROR, "Could not allocated memory for background job parameter.");
		fFailed = TRUE;
	}

	if (!fFailed)
	{
		memset(lpResult, 0, sizeof(*lpResult));
		lpResult->lpProgressData = lpProgressData;
		lpResult->lpRemotePath = lpSourcePath;
	}

	if (!fFailed)
	{
		lpResult->hLocalFile = CreateFile(lpDestinationPath, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);
		if (lpResult->hLocalFile == INVALID_HANDLE_VALUE)
		{
			FormatMessageBox(hWnd, MB_ICONERROR, "Could not open file: %s", GetLastErrorText());
			fFailed = TRUE;
		}
	}

	if (fFailed)
		SafeDestroyTransferFileJobProcParam(&lpResult);

	return lpResult;
}

static BOOL RecvFileChunk(struct CommDev *lpDevice, struct TransferFileJobProcParam *lpData, LPWORD lpBytesRead)
{
	BOOL fResult = FALSE;

	struct ReadFileRequest request = { CMD_READ_FILE };
	request.wFileHandle = lpData->wRemoteFileHandle;

	struct ReadFileResponse response;

	WORD wResponseSize = sizeof(response);
	if (ProtocolTransaction(lpDevice, &request, sizeof(request), &response, &wResponseSize))
	{
		if (response.bDosStatusCode == DOS_CODE_SUCCESS)
		{
			DWORD dwBytesWritten;
			if (WriteFile(lpData->hLocalFile, response.FileData, response.wBytesRead, &dwBytesWritten, NULL))
			{
				if (dwBytesWritten == response.wBytesRead)
				{
					*lpBytesRead = response.wBytesRead;
					fResult = TRUE;
				}
				else
				{
					TRACE("Short WriteFile, %lu bytes instead of %i\n", dwBytesWritten, response.wBytesRead);
				}
			}
			else
			{
				TRACE("WriteFile failed: %lu\n", GetLastError());
			}
		}
		else
		{
			TRACE("Negative DOS status code: 0x%.2u\n", response.bDosStatusCode);
		}
	}

	return fResult;
}

static WORD RecvFileJobProc(struct CommDev *lpDevice, BYTE bAction, LPVOID lpParam)
{
	ASSERT(lpParam != NULL);

	WORD wResult = BGJOB_RESULT_FAILURE;
	struct TransferFileJobProcParam *lpData = lpParam;

	if (bAction == BGJOB_ACTION_EXECUTE)
	{
		if (!lpData->fRemoteFileOpen)
		{
			if (OpenRemoteFile(lpDevice, lpData->lpRemotePath, PRM_OPEN_FILE_READ, &lpData->wRemoteFileHandle))
			{
				// OK, continue next time with next step
				lpData->fRemoteFileOpen = TRUE;
				wResult = BGJOB_RESULT_RUNNING;
			}
		}
		else if (!lpData->fSourceSizeKnown)
		{
			if (GetRemoteFileSize(lpDevice, lpData->wRemoteFileHandle, &lpData->dwTotalBytes))
			{
				// OK, continue next time with next step
				lpData->fSourceSizeKnown = TRUE;
				wResult = BGJOB_RESULT_RUNNING;
			}
			else
			{
				CloseRemoteFile(lpDevice, lpData->wRemoteFileHandle);
				wResult = BGJOB_RESULT_FAILURE;
			}
		}
		else if (lpData->dwBytesProcessed < lpData->dwTotalBytes)
		{
			WORD wBytesRead;
			if (RecvFileChunk(lpDevice, lpData, &wBytesRead))
			{
				lpData->dwBytesProcessed += wBytesRead;
				EnterCriticalSection(&lpData->lpProgressData->csLock);
				lpData->lpProgressData->dwBytesProcessed = lpData->dwBytesProcessed;
				lpData->lpProgressData->dwTotalBytes = lpData->dwTotalBytes;
				LeaveCriticalSection(&lpData->lpProgressData->csLock);
				wResult = BGJOB_RESULT_RUNNING;
			}
			else
			{
				CloseRemoteFile(lpDevice, lpData->wRemoteFileHandle);
				wResult = BGJOB_RESULT_FAILURE;
			}
		}
		else if (lpData->fRemoteFileOpen)
		{
			if (CloseRemoteFile(lpDevice, lpData->wRemoteFileHandle))
			{
				// Last step completed
				wResult = BGJOB_RESULT_SUCCESS;
			}
			lpData->fRemoteFileOpen = FALSE;
		}
	}
	else if (bAction == BGJOB_ACTION_CANCEL)
	{
		if (lpData->fRemoteFileOpen)
			CloseRemoteFile(lpDevice, lpData->wRemoteFileHandle);
	}

	return wResult;
}

BOOL RecvFile(HWND hWnd, struct CommDev *lpDevice, LPCSTR lpSourcePath, LPCSTR lpDestinationPath)
{
	BOOL fFailed = FALSE;

	struct TransferFileProgress Progress = { 0 };
	InitializeCriticalSection(&Progress.csLock);

	struct TransferDlgData DlgData = { 0 };
	DlgData.lpProgressData = &Progress;
	DlgData.lpUpdateProgressProc = TransferDlg_UpdateProgress;
	DlgData.lpSource = lpSourcePath;
	DlgData.lpDestination = lpDestinationPath;

	struct TransferFileJobProcParam *lpParam = CreateRecvFileJobProcParam(hWnd, lpSourcePath, lpDestinationPath, &Progress);
	if (lpParam)
	{
		struct BackgroundJob *lpJob = CreateBackgroundJob(lpDevice, RecvFileJobProc, lpParam);
		if (lpJob)
		{
			DlgData.lpJob = lpJob;

			HINSTANCE hInst = (HINSTANCE) GetWindowLongPtr(hWnd, GWLP_HINSTANCE);
			LPCSTR lpTemplate = MAKEINTRESOURCE(IDD_TRANSFERDLG);
			if (DialogBoxParam(hInst, lpTemplate, hWnd, TransferDlgProc, (LPARAM) &DlgData) == BGJOB_RESULT_SUCCESS)
			{
				// Nothing to do
			}
			else
			{
				TRACE("Job ended unhappy\n");
				fFailed = TRUE;
			}

			DestroyBackgroundJob(lpJob);
		}
		else
		{
			TRACE("Could not create background job\n");
			fFailed = TRUE;
		}

		DestroyTransferFileJobProcParam(lpParam);
	}
	else
	{
		fFailed = TRUE;
	}

	DeleteCriticalSection(&Progress.csLock);
	return !fFailed;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Specific part for sending file to target

static struct TransferFileJobProcParam *CreateSendFileJobProcParam(HWND hWnd, LPCSTR lpSourcePath,
	LPCSTR lpDestinationPath, struct TransferFileProgress *lpProgressData)
{
	BOOL fFailed = FALSE;

	struct TransferFileJobProcParam *lpResult = MALLOC(sizeof(*lpResult));
	if (lpResult == NULL)
	{
		FormatMessageBox(hWnd, MB_ICONERROR, "Could not allocated memory for background job parameter.");
		fFailed = TRUE;
	}

	if (!fFailed)
	{
		memset(lpResult, 0, sizeof(*lpResult));
		lpResult->lpProgressData = lpProgressData;
		lpResult->lpRemotePath = lpDestinationPath;
	}

	if (!fFailed)
	{
		lpResult->hLocalFile = CreateFile(lpSourcePath, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
		if (lpResult->hLocalFile == INVALID_HANDLE_VALUE)
		{
			FormatMessageBox(hWnd, MB_ICONERROR, "Could not open file: %s", GetLastErrorText());
			fFailed = TRUE;
		}
	}

	if (!fFailed)
	{
		lpResult->dwTotalBytes = GetFileSize(lpResult->hLocalFile, NULL);
		if (lpResult->dwTotalBytes != INVALID_FILE_SIZE)
		{
			lpResult->fSourceSizeKnown = TRUE;
		}
		else
		{
			FormatMessageBox(hWnd, MB_ICONERROR, "Could not get file size: %s", GetLastErrorText());
			fFailed = TRUE;
		}
	}

	if (fFailed)
		SafeDestroyTransferFileJobProcParam(&lpResult);

	return lpResult;
}

static BOOL SendFileChunk(struct CommDev *lpDevice, struct TransferFileJobProcParam *lpData, LPWORD lpBytesSent)
{
	BOOL fFailed = FALSE;

	struct WriteFileRequest request = { CMD_WRITE_FILE };
	request.wFileHandle = lpData->wRemoteFileHandle;

	struct WriteFileResponse response;

	DWORD dwBytesRead;
	if (ReadFile(lpData->hLocalFile, request.FileData, sizeof(request.FileData), &dwBytesRead, NULL) == FALSE)
	{
		TRACE("Could not read from file: %lu\n", GetLastError());
		fFailed = TRUE;
	}

	if (!fFailed)
	{
		request.wBytesToWrite = (WORD) dwBytesRead;
		WORD wResponseSize = sizeof(response);
		fFailed = ProtocolTransaction(lpDevice, &request, sizeof(request), &response, &wResponseSize) == FALSE;
	}

	if (!fFailed)
	{
		if (response.bDosStatusCode != DOS_CODE_SUCCESS)
		{
			TRACE("Negative DOS status code: 0x%.2u\n", response.bDosStatusCode);
			fFailed = TRUE;
		}
	}

	if (!fFailed)
	{
		if (response.wBytesWritten != request.wBytesToWrite)
		{
			TRACE("Short write on target (%u instead of %u)\n", response.wBytesWritten, request.wBytesToWrite);
			fFailed = TRUE;
		}
	}

	if (!fFailed)
	{
		*lpBytesSent = response.wBytesWritten;
	}

	return !fFailed;
}

static WORD SendFileJobProc(struct CommDev *lpDevice, BYTE bAction, LPVOID lpParam)
{
	ASSERT(lpParam != NULL);

	WORD wResult = BGJOB_RESULT_FAILURE;

	struct TransferFileJobProcParam *lpData = lpParam;

	if (bAction == BGJOB_ACTION_EXECUTE)
	{
		if (!lpData->fRemoteFileOpen)
		{
			if (OpenRemoteFile(lpDevice, lpData->lpRemotePath, PRM_OPEN_FILE_WRITE, &lpData->wRemoteFileHandle))
			{
				// OK, continue next time with next step
				lpData->fRemoteFileOpen = TRUE;
				wResult = BGJOB_RESULT_RUNNING;
			}
		}
		else if (lpData->dwBytesProcessed < lpData->dwTotalBytes)
		{
			WORD wBytesSent;
			if (SendFileChunk(lpDevice, lpData, &wBytesSent))
			{
				lpData->dwBytesProcessed += wBytesSent;
				EnterCriticalSection(&lpData->lpProgressData->csLock);
				lpData->lpProgressData->dwBytesProcessed = lpData->dwBytesProcessed;
				lpData->lpProgressData->dwTotalBytes = lpData->dwTotalBytes;
				LeaveCriticalSection(&lpData->lpProgressData->csLock);
				wResult = BGJOB_RESULT_RUNNING;
			}
			else
			{
				CloseRemoteFile(lpDevice, lpData->wRemoteFileHandle);
				wResult = BGJOB_RESULT_FAILURE;
			}
		}
		else if (lpData->fRemoteFileOpen)
		{
			if (CloseRemoteFile(lpDevice, lpData->wRemoteFileHandle))
			{
				// Last step completed
				wResult = BGJOB_RESULT_SUCCESS;
			}
			lpData->fRemoteFileOpen = FALSE;
		}
	}
	else if (bAction == BGJOB_ACTION_CANCEL)
	{
		if (lpData->fRemoteFileOpen)
			CloseRemoteFile(lpDevice, lpData->wRemoteFileHandle);
	}

	return wResult;
}

BOOL SendFile(HWND hWnd, struct CommDev *lpDevice, LPCSTR lpSourcePath, LPCSTR lpDestinationPath)
{
	BOOL fFailed = FALSE;

	struct TransferFileProgress Progress = { 0 };
	InitializeCriticalSection(&Progress.csLock);

	struct TransferDlgData DlgData = { 0 };
	DlgData.lpProgressData = &Progress;
	DlgData.lpUpdateProgressProc = TransferDlg_UpdateProgress;
	DlgData.lpSource = lpSourcePath;
	DlgData.lpDestination = lpDestinationPath;

	struct TransferFileJobProcParam *lpParam = CreateSendFileJobProcParam(hWnd, lpSourcePath, lpDestinationPath, &Progress);
	if (lpParam)
	{
		Progress.dwTotalBytes = lpParam->dwTotalBytes;
		struct BackgroundJob *lpJob = CreateBackgroundJob(lpDevice, SendFileJobProc, lpParam);
		if (lpJob)
		{
			DlgData.lpJob = lpJob;

			HINSTANCE hInst = (HINSTANCE) GetWindowLongPtr(hWnd, GWLP_HINSTANCE);
			LPCSTR lpTemplate = MAKEINTRESOURCE(IDD_TRANSFERDLG);
			if (DialogBoxParam(hInst, lpTemplate, hWnd, TransferDlgProc, (LPARAM) &DlgData) == BGJOB_RESULT_SUCCESS)
			{
				// Nothing to do
			}
			else
			{
				TRACE("Job ended unhappy\n");
				fFailed = TRUE;
			}

			DestroyBackgroundJob(lpJob);
		}
		else
		{
			TRACE("Could not create background job\n");
			fFailed = TRUE;
		}

		DestroyTransferFileJobProcParam(lpParam);
	}
	else
	{
		fFailed = TRUE;
	}

	DeleteCriticalSection(&Progress.csLock);
	return !fFailed;
}
