#include <stdio.h>
#include <windows.h>
#include <commctrl.h>

#include "commproto.h"
#include "bgjob.h"
#include "protodef.h"
#include "resource.h"
#include "guiutils.h"
#include "transferdlg.h"

#include "debug.h"

#include "transferimage.h"


///////////////////////////////////////////////////////////////////////////////////////////////////
// Generic part for both directions (read and write)

struct DrivePosition {
	BYTE bTrack;
	BYTE bSide;
	BYTE bSector;
};

struct TransferImageProgress {
	CRITICAL_SECTION csLock;
	struct DrivePosition dpCurrPos;
	WORD wCurrStep;
	WORD wTotalSteps;
};

#define DEF_BUFFER_SIZE 100

static void TransferDlg_UpdateProgress(HWND hDlg, struct TransferDlgData *lpData)
{
	// Keep lock as short as possible
	struct TransferImageProgress *lpProgressData = lpData->lpProgressData;
	EnterCriticalSection(&lpProgressData->csLock);
	struct DrivePosition dpCurrPos = lpProgressData->dpCurrPos;
	WORD wCurrStep = lpProgressData->wCurrStep;
	WORD wTotalSteps = lpProgressData->wTotalSteps;
	LeaveCriticalSection(&lpProgressData->csLock);

	char szBuffer[DEF_BUFFER_SIZE];
	int iPercent = (int) (wTotalSteps > 0 ? (wCurrStep * 100.0 / wTotalSteps + 0.5) : 0);

	if (wCurrStep > 0)
	{
		DWORD dwTicksPassed = GetTickCount() - lpData->dwBeginOfTransfer;
		double dTicksPerStep = dwTicksPassed / (double) wCurrStep;
		DWORD dwTicksLeft = (DWORD) (dTicksPerStep * wTotalSteps) - dwTicksPassed;
		int iSecondsLeft = (dwTicksLeft + 750) / 1000; // Staying too long at "00:00" looks stupid

		sprintf_s(szBuffer, sizeof(szBuffer), "Track %i, side %i, sector %i (%i%%), %i:%.2i left",
			dpCurrPos.bTrack, dpCurrPos.bSide, dpCurrPos.bSector, iPercent, iSecondsLeft / 60, iSecondsLeft % 60);
	}
	else
	{
		sprintf_s(szBuffer, sizeof(szBuffer), "Track %i, side %i, sector %i (%i%%)",
			dpCurrPos.bTrack, dpCurrPos.bSide, dpCurrPos.bSector, iPercent);
	}

	SetWindowText(GetDlgItem(hDlg, IDC_STATUS), szBuffer);
	SendMessage(lpData->hProgressBar, PBM_SETRANGE, 0, MAKELPARAM(0, wTotalSteps));
	SendMessage(lpData->hProgressBar, PBM_SETPOS, wCurrStep, 0);
}

struct TransferImageJobProcParam {
	BYTE bDriveIndex;
	const struct DriveInfo *lpInfo;
	HANDLE hFile;
	struct DrivePosition dpCurrPos;
	struct TransferImageProgress *lpProgressData;
};

static void SafeDestroyTransferImageJobProcParam(struct TransferImageJobProcParam **lpParamPtr)
{
	if (*lpParamPtr)
	{
		if ((*lpParamPtr)->hFile != INVALID_HANDLE_VALUE)
			CloseHandle((*lpParamPtr)->hFile);
		FREE(*lpParamPtr);
		*lpParamPtr = NULL;
	}
}

static void DestroyTransferImageJobProcParam(struct TransferImageJobProcParam *lpParam)
{
	ASSERT(lpParam);
	SafeDestroyTransferImageJobProcParam(&lpParam);
}

static BOOL AdvanceToNextSector(struct DrivePosition *lpOldPos, const struct DriveInfo *lpInfo)
{
	BOOL fFinished = FALSE;

	struct DrivePosition dpNewPos = *lpOldPos;
	if (++dpNewPos.bSector >= lpInfo->bNumberOfSectors)
	{
		dpNewPos.bSector = 0;
		if (++dpNewPos.bSide >= lpInfo->bNumberOfSides)
		{
			dpNewPos.bSide = 0;
			if (++dpNewPos.bTrack >= lpInfo->bNumberOfTracks)
			{
				fFinished = TRUE;
			}
		}
	}

	if (!fFinished)
		*lpOldPos = dpNewPos;

	return !fFinished;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Specific part for reading image from target

static struct TransferImageJobProcParam *CreateReadImageJobProcParam(HWND hWnd, BYTE bDriveIndex,
	const struct DriveInfo *lpInfo, LPCSTR lpPath, struct TransferImageProgress *lpProgressData)
{
	BOOL fFailed = FALSE;

	struct TransferImageJobProcParam *lpResult = MALLOC(sizeof(*lpResult));
	if (lpResult == NULL)
	{
		FormatMessageBox(hWnd, MB_ICONERROR, "Could not allocated memory for background job parameter.");
		fFailed = TRUE;
	}

	if (!fFailed)
	{
		lpResult->bDriveIndex = bDriveIndex;
		lpResult->lpInfo = lpInfo;
		lpResult->dpCurrPos.bTrack = 0;
		lpResult->dpCurrPos.bSide = 0;
		lpResult->dpCurrPos.bSector = 0;
		lpResult->lpProgressData = lpProgressData;
	}

	if (!fFailed)
	{
		lpResult->hFile = CreateFile(lpPath, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);
		if (lpResult->hFile == INVALID_HANDLE_VALUE)
		{
			FormatMessageBox(hWnd, MB_ICONERROR, "Could not open file: %s", GetLastErrorText());
			fFailed = TRUE;
		}
	}

	if (fFailed)
		SafeDestroyTransferImageJobProcParam(&lpResult);

	return lpResult;
}

static BOOL ReadSector(struct CommDev *lpDevice, BYTE bDriveIndex, struct DrivePosition dpPos, HANDLE hFile)
{
	BOOL fResult = FALSE;

	struct ReadSectorRequest request = { CMD_READ_SECTOR };
	request.bDriveIndex = bDriveIndex;
	request.bTrack = dpPos.bTrack;
	request.bSide = dpPos.bSide;
	request.bSector = dpPos.bSector;

	struct ReadSectorResponse response;

	WORD wResponseSize = sizeof(response);
	if (ProtocolTransaction(lpDevice, &request, sizeof(request), &response, &wResponseSize))
	{
		if (response.bBiosStatusCode == BIOS_CODE_SUCCESS)
		{
			DWORD dwBytesWritten;
			if (WriteFile(hFile, response.SectorData, SECTOR_SIZE, &dwBytesWritten, NULL))
			{
				if (dwBytesWritten == SECTOR_SIZE)
					fResult = TRUE;
				else
					TRACE("Short WriteFile, %lu bytes instead of %i\n", dwBytesWritten, SECTOR_SIZE);
			}
			else
			{
				TRACE("WriteFile failed: %lu\n", GetLastError());
			}
		}
		else
		{
			TRACE("Negative BIOS status code: 0x%.2u\n", response.bBiosStatusCode);
		}
	}

	return fResult;
}

static WORD ReadImageJobProc(struct CommDev *lpDevice, BYTE bAction, LPVOID lpParam)
{
	ASSERT(lpParam != NULL);

	WORD wResult = BGJOB_RESULT_FAILURE;
	struct TransferImageJobProcParam *lpData = lpParam;

	if (bAction == BGJOB_ACTION_EXECUTE)
	{
		if (ReadSector(lpDevice, lpData->bDriveIndex, lpData->dpCurrPos, lpData->hFile))
		{
			wResult = AdvanceToNextSector(&lpData->dpCurrPos, lpData->lpInfo) ? BGJOB_RESULT_RUNNING : BGJOB_RESULT_SUCCESS;

			EnterCriticalSection(&lpData->lpProgressData->csLock);
			lpData->lpProgressData->dpCurrPos = lpData->dpCurrPos;
			lpData->lpProgressData->wCurrStep++;
			LeaveCriticalSection(&lpData->lpProgressData->csLock);
		}
		else
		{
			wResult = BGJOB_RESULT_FAILURE;
		}
	}

	return wResult;
}

BOOL ReadImage(HWND hWnd, struct CommDev *lpDevice, BYTE bDriveIndex, const struct DriveInfo *lpInfo, LPCSTR lpPath)
{
	BOOL fFailed = FALSE;

	struct TransferImageProgress Progress = { 0 };
	Progress.wTotalSteps = lpInfo->bNumberOfTracks * lpInfo->bNumberOfSides * lpInfo->bNumberOfSectors;
	InitializeCriticalSection(&Progress.csLock);

	struct TransferDlgData DlgData = { 0 };
	DlgData.lpProgressData = &Progress;
	DlgData.lpUpdateProgressProc = TransferDlg_UpdateProgress;

	char szBuffer[MAX_PATH];
	sprintf_s(szBuffer, sizeof(szBuffer), "Drive %i (%s)", bDriveIndex, GetDriveTypeString(lpInfo->bDriveType));
	DlgData.lpSource = szBuffer;
	DlgData.lpDestination = lpPath;

	struct TransferImageJobProcParam *lpParam = CreateReadImageJobProcParam(hWnd, bDriveIndex, lpInfo, lpPath, &Progress);
	if (lpParam)
	{
		struct BackgroundJob *lpJob = CreateBackgroundJob(lpDevice, ReadImageJobProc, lpParam);
		if (lpJob)
		{
			DlgData.lpJob = lpJob;

			HINSTANCE hInst = (HINSTANCE) GetWindowLongPtr(hWnd, GWLP_HINSTANCE);
			LPCSTR lpTemplate = MAKEINTRESOURCE(IDD_TRANSFERDLG);
			if (DialogBoxParam(hInst, lpTemplate, hWnd, TransferDlgProc, (LPARAM) &DlgData) == BGJOB_RESULT_SUCCESS)
			{
				// Nothing to do
			}
			else
			{
				TRACE("Job ended unhappy\n");
				fFailed = TRUE;
			}

			DestroyBackgroundJob(lpJob);
		}
		else
		{
			TRACE("Could not create background job\n");
			fFailed = TRUE;
		}

		DestroyTransferImageJobProcParam(lpParam);
	}
	else
	{
		fFailed = TRUE;
	}

	DeleteCriticalSection(&Progress.csLock);
	return !fFailed;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Specific part for writing image to target

static struct TransferImageJobProcParam *CreateWriteImageJobProcParam(HWND hWnd, BYTE bDriveIndex,
	const struct DriveInfo *lpInfo, LPCSTR lpPath, struct TransferImageProgress *lpProgressData)
{
	BOOL fFailed = FALSE;

	struct TransferImageJobProcParam *lpResult = MALLOC(sizeof(*lpResult));
	if (lpResult == NULL)
	{
		FormatMessageBox(hWnd, MB_ICONERROR, "Could not allocated memory for background job parameter.");
		fFailed = TRUE;
	}

	if (!fFailed)
	{
		lpResult->bDriveIndex = bDriveIndex;
		lpResult->lpInfo = lpInfo;
		lpResult->dpCurrPos.bTrack = 0;
		lpResult->dpCurrPos.bSide = 0;
		lpResult->dpCurrPos.bSector = 0;
		lpResult->lpProgressData = lpProgressData;
	}

	if (!fFailed)
	{
		lpResult->hFile = CreateFile(lpPath, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
		if (lpResult->hFile == INVALID_HANDLE_VALUE)
		{
			FormatMessageBox(hWnd, MB_ICONERROR, "Could not open file: %s", GetLastErrorText());
			fFailed = TRUE;
		}
	}

	if (!fFailed)
	{
		DWORD dwActualSize = GetFileSize(lpResult->hFile, NULL);
		DWORD dwExpectedSize = lpInfo->bNumberOfTracks * lpInfo->bNumberOfSides * lpInfo->bNumberOfSectors * SECTOR_SIZE;
		if (dwActualSize == INVALID_FILE_SIZE)
		{
			FormatMessageBox(hWnd, MB_ICONERROR, "Could not get file size: %s", GetLastErrorText());
			fFailed = TRUE;
		}
		else if (dwActualSize != dwExpectedSize)
		{
			int iResult = FormatMessageBox(hWnd, MB_ICONQUESTION | MB_YESNO | MB_DEFBUTTON2,
				"Unexpected file size for this type of drive!\n\n"
				"Expected: %u bytes (%uk)\n"
				"Actual: %u bytes (%uk)\n\n"
				"Do you want to continue using this file?",
				dwExpectedSize, dwExpectedSize/1024, dwActualSize, dwActualSize/1024);
			fFailed = iResult != IDYES;
		}
	}

	if (fFailed)
		SafeDestroyTransferImageJobProcParam(&lpResult);

	return lpResult;
}

static BOOL WriteSector(struct CommDev *lpDevice, BYTE bDriveIndex, struct DrivePosition dpPos, HANDLE hFile)
{
	BOOL fFailed = FALSE;

	struct WriteSectorRequest request = { CMD_WRITE_SECTOR };
	request.bDriveIndex = bDriveIndex;
	request.bTrack = dpPos.bTrack;
	request.bSide = dpPos.bSide;
	request.bSector = dpPos.bSector;

	struct WriteSectorResponse response;

	DWORD dwBytesRead;
	if (ReadFile(hFile, request.SectorData, SECTOR_SIZE, &dwBytesRead, NULL) == FALSE)
	{
		TRACE("Could not read from file: %lu\n", GetLastError());
		fFailed = TRUE;
	}

	if (!fFailed)
	{
		if (dwBytesRead != SECTOR_SIZE)
		{
			TRACE("Short read: got %lu instead of %i bytes\n", dwBytesRead, SECTOR_SIZE);
			fFailed = TRUE;
		}
	}

	if (!fFailed)
	{
		WORD wResponseSize = sizeof(response);
		fFailed = ProtocolTransaction(lpDevice, &request, sizeof(request), &response, &wResponseSize) == FALSE;
	}

	if (!fFailed)
	{
		if (response.bBiosStatusCode != BIOS_CODE_SUCCESS)
		{
			TRACE("Negative BIOS status code: 0x%.2u\n", response.bBiosStatusCode);
			fFailed = TRUE;
		}
	}

	return !fFailed;
}

static WORD WriteImageJobProc(struct CommDev *lpDevice, BYTE bAction, LPVOID lpParam)
{
	ASSERT(lpParam != NULL);

	WORD wResult = BGJOB_RESULT_FAILURE;

	if (bAction == BGJOB_ACTION_EXECUTE)
	{
		struct TransferImageJobProcParam *lpData = lpParam;

		if (WriteSector(lpDevice, lpData->bDriveIndex, lpData->dpCurrPos, lpData->hFile))
		{
			wResult = AdvanceToNextSector(&lpData->dpCurrPos, lpData->lpInfo) ? BGJOB_RESULT_RUNNING : BGJOB_RESULT_SUCCESS;

			EnterCriticalSection(&lpData->lpProgressData->csLock);
			lpData->lpProgressData->dpCurrPos = lpData->dpCurrPos;
			lpData->lpProgressData->wCurrStep++;
			LeaveCriticalSection(&lpData->lpProgressData->csLock);
		}
		else
		{
			wResult = BGJOB_RESULT_FAILURE;
		}
	}

	return wResult;
}

BOOL WriteImage(HWND hWnd, struct CommDev *lpDevice, BYTE bDriveIndex, const struct DriveInfo *lpInfo, LPCSTR lpPath)
{
	BOOL fFailed = FALSE;

	struct TransferImageProgress Progress = { 0 };
	Progress.wTotalSteps = lpInfo->bNumberOfTracks * lpInfo->bNumberOfSides * lpInfo->bNumberOfSectors;
	InitializeCriticalSection(&Progress.csLock);

	struct TransferDlgData DlgData = { 0 };
	DlgData.lpProgressData = &Progress;
	DlgData.lpUpdateProgressProc = TransferDlg_UpdateProgress;

	char szBuffer[MAX_PATH];
	sprintf_s(szBuffer, sizeof(szBuffer), "Drive %i (%s)", bDriveIndex, GetDriveTypeString(lpInfo->bDriveType));
	DlgData.lpSource = lpPath;
	DlgData.lpDestination = szBuffer;

	struct TransferImageJobProcParam *lpParam = CreateWriteImageJobProcParam(hWnd, bDriveIndex, lpInfo, lpPath, &Progress);
	if (lpParam)
	{
		struct BackgroundJob *lpJob = CreateBackgroundJob(lpDevice, WriteImageJobProc, lpParam);
		if (lpJob)
		{
			DlgData.lpJob = lpJob;

			HINSTANCE hInst = (HINSTANCE) GetWindowLongPtr(hWnd, GWLP_HINSTANCE);
			LPCSTR lpTemplate = MAKEINTRESOURCE(IDD_TRANSFERDLG);
			if (DialogBoxParam(hInst, lpTemplate, hWnd, TransferDlgProc, (LPARAM) &DlgData) == BGJOB_RESULT_SUCCESS)
			{
				// Nothing to do
			}
			else
			{
				TRACE("Job ended unhappy\n");
				fFailed = TRUE;
			}

			DestroyBackgroundJob(lpJob);
		}
		else
		{
			TRACE("Could not create background job\n");
			fFailed = TRUE;
		}

		DestroyTransferImageJobProcParam(lpParam);
	}
	else
	{
		fFailed = TRUE;
	}

	DeleteCriticalSection(&Progress.csLock);
	return !fFailed;
}
