;******************************************************************************
;
; Filename:     toolbox.asm
;
; Author:       Felix Opatz <felix@zotteljedi.de>
;
; Description:  Server that allows clients to manipulate the target system.
;
;               The toolbox is loaded to the target system and interacts with
;               clients via the COM port connection.
;               Functions:
;               - Exit toolbox with custom exit code
;               - Get toolbox version (for basic "are u there?")
;               - Get disk drive information (for enumeration)
;               - Read sector from disk drive
;               - Write sector to disk drive
;               - Open/create file in target file system
;               - Get size of open file
;               - Read data from open file
;               - Write data to open file
;               - Close open file
;
;******************************************************************************

;******************************************************************************
; Origin address of code
;******************************************************************************
	org	LOAD_ADDR		; Origin for user programs


;******************************************************************************
; Constants
;******************************************************************************
CMD_EXIT		equ	0	; Function "exit with custom code"
CMD_GET_VERSION		equ	1	; Function "get toolbox version"
CMD_GET_DRIVE_INFO	equ	2	; Function "get drive information"
CMD_READ_SECTOR		equ	3	; Function "read sector from drive"
CMD_WRITE_SECTOR	equ	4	; Function "write sector to drive"
CMD_OPEN_FILE		equ	5	; Function "open file"
CMD_GET_FILE_SIZE	equ	6	; Function "get file size"
CMD_READ_FILE		equ	7	; Function "read from file"
CMD_WRITE_FILE		equ	8	; Function "write to file"
CMD_CLOSE_FILE		equ	9	; Function "close file"

PRM_OPEN_FILE_READ	equ	0	; Access mode "read"
PRM_OPEN_FILE_WRITE	equ	1	; Access mode "write" (create)

REQ_BUFFER_SIZE		equ	1024	; Size of request buffer in bytes
RES_BUFFER_SIZE		equ	1024	; Size of response buffer in bytes
RES_FLAG_SUCCESS	equ	0x40	; Flag for successful requests
RES_FLAG_FAILURE	equ	0x80	; Flag for failed requests
REQ_MIN_LENGTH		equ	1	; Minimum length of a request


;******************************************************************************
; Variables
;******************************************************************************
section .bss
; Layout of request messages:
; - 1 byte command code
; - n bytes of command specific data
Request			resb	REQ_BUFFER_SIZE

; Layout of response messages:
; - 1 byte repsonse code (= request code + response flags)
; - n bytes of command specific data
Response		resb	RES_BUFFER_SIZE


section .text
;******************************************************************************
;
; Function:     ToolboxMain, program entry point
;
; Parameters:   [bp+4],2: COM port to receive on
;
; Returns:      AX: exit code of toolbox via CMD_EXIT
;
;******************************************************************************
ToolboxMain:
	push	bp
	mov	bp, sp
	push	bx

	call	PrintBanner
.loop:
	; Read request from COM port
	mov	bx, REQ_BUFFER_SIZE
	push	bx			; Parameter 3: request buffer size
	mov	bx, Request
	push	bx			; Parameter 2: request buffer
	mov	bx, [bp + 4]
	push	bx			; Parameter 1: COM port to read from
	call	ReadRequest		; AX: number of bytes read
	cmp	ax, REQ_MIN_LENGTH	; Check for a short read (timeout)
	jb	.loop

	; Special handling for CMD_EXIT
	cmp	byte [Request], CMD_EXIT
	je	.ProcessExit

	; Regular handling of all other requests
	mov	bx, RES_BUFFER_SIZE
	push	bx			; Parameter 4: response buffer size
	mov	bx, Response
	push	bx			; Parameter 3: response buffer
	push	ax			; Parameter 2: request length
	mov	bx, Request
	push	bx			; Parameter 1: request buffer
	call	ProcessRequest		; AX: number of bytes in response

	; Write response to COM port
	push	ax			; Parameter 3: number of bytes to send
	mov	bx, Response
	push	bx			; Parameter 2: response buffer
	mov	bx, [bp + 4]
	push	bx			; Parameter 1: COM port to write to
	call	WriteResponse		; No return value

	jmp	.loop
.ProcessExit:
	cmp	ax, 2			; Check if an exit code was specified
	jae	.end			; Yes: then go the normal way
	mov	byte [Request + 1], 0	; No: fake a value of 0
.end:
	mov	al, [Request + 1]	; 2. byte of CMD_EXIT is return code
	mov	ah, 0			; Make sure there is no garbage in AH

	pop	bx
	pop	bp
	ret	2			; Return to initial loader


;******************************************************************************
;
; Function:     PrintBanner
;
; Description:  Prints an info message that the toolbox was started.
;               Works currently only with single digit numbers for major and
;               minor version.
;
; Parameters:   None.
;
; Returns:      Void.
;
;******************************************************************************
PrintBanner:
	push	ax
	push	dx

	mov	dx, banner_1		; Part 1 of banner text
	mov	ah, 0x09		; Function "write string"
	int	0x21			; Trigger DOS interrupt
	mov	ah, 0x02		; Function "write character"
	mov	dl, VERSION_MAJOR	; Get major version number
	add	dl, '0'			; Make binary value to digit
	int	0x21			; Trigger DOS interrupt
	mov	dl, '.'			; Write dot between major/minor
	int	0x21			; Trigger DOS interrupt
	mov	dl, VERSION_MINOR	; Get minor version number
	add	dl, '0'			; Make binary value to digit
	int	0x21			; Trigger DOS interrupt
	mov	dx, banner_2		; Part 2 of banner text
	mov	ah, 0x09		; Function "write string"
	int	0x21			; Trigger DOS interrupt

	pop	dx
	pop	ax
	ret


;******************************************************************************
;
; Function:     ReadRequest
;
; Description:  Reads a request from the specified COM port.
;
;               Reads 2 bytes to form the number of bytes in packet.
;               Then reads that number of bytes into buffer.
;               When reading fails, the call returns with what it has.
;
; Parameters:   [bp+4],2: COM port to receive on
;               [bp+6],2: Buffer to store request in.
;               [bp+8],2: Size of buffer
;
; Returns:      AX: number of bytes stored in request buffer.
;
;******************************************************************************
ReadRequest:
	push	bp
	mov	bp, sp
	pushf
	push	cx
	push	dx
	push	di

	cld				; Increase DI when writing
	mov	dx, [bp + 4]		; COM port to read from
	mov	di, [bp + 6]		; Pointer to buffer
.readSizeLow:
	mov	ah, 0x02		; Function "receive character"
	int	0x14			; Trigger BIOS interrupt
	test	ah, 0x80		; Test for failure flag
	jnz	.done			; If failed, then stop
	mov	cl, al			; Store byte as low byte
.readSizeHigh:
	mov	ah, 0x02		; Function "receive character"
	int	0x14			; Trigger BIOS interrupt
	test	ah, 0x80		; Test for failure flag
	jnz	.done			; If failed, then stop
	mov	ch, al			; Store byte as high byte
.getMinimum:
	cmp	cx, [bp + 8]		; Compare expected size/buffer size
	jbe	.sizeOK
	mov	cx, [bp + 8]		; Take buffer size instead
.sizeOK:
	cmp	cx, 0			; Check if it is a null read,
	je	.done			;   then skip it
.receive:
	mov	ah, 0x02		; Function "receive character"
	int	0x14			; Trigger BIOS interrupt
	test	ah, 0x80		; Test for failure flag
	jnz	.done			; If failed, then stop
	stosb				; Store received byte
	loop	.receive		; Loop for next byte
.done:
	mov	ax, di			; Get current pointer
	sub	ax, [bp + 6]		; Get difference to origin
.exit:
	pop	di
	pop	dx
	pop	cx
	popf
	pop	bp
	ret	6


;******************************************************************************
;
; Function:     ProcessRequest
;
; Description:  Process a received request.
;
; Parameters:   [bp+ 4],2: Request
;               [bp+ 6],2: Size of request
;               [bp+ 8],2: Buffer to receive response
;               [bp+10],2: Size of response buffer
;
; Returns:      AX: number of bytes stored in response buffer.
;
;******************************************************************************
ProcessRequest:
	push	bp
	mov	bp, sp
	push	bx
	push	cx
	push	si
	
	mov	si, .table		; Set SI to lookup table
	mov	bx, [bp + 4]		; Get pointer to request
	mov	al, [bx]		; Copy request code into AL
	mov	ah, 0			; Make AH zero to match entire word
	mov	cx, .count		; Loop over all entries
.loop:
	cmp	ax, [si]		; Check if request code matches
	je	.found
	add	si, 4			; Advance SI to next table entry
	loop	.loop
	mov	bx, ProcessInvalidCommand	; Code was not found
	jmp	.exec
.found:
	mov	bx, [si + 2]		; Get address of process function

.exec:
	mov	ax, [bp + 10]
	push	ax			; Parameter 4: response buffer size
	mov	ax, [bp + 8]
	push	ax			; Parameter 3: response buffer
	mov	ax, [bp + 6]
	push	ax			; Parameter 2: request length
	mov	ax, [bp + 4]
	push	ax			; Parameter 1: request buffer
	call	bx			; Call process function

	pop	si
	pop	cx
	pop	bx
	pop	bp
	ret	8

		;	Request code        Process function
.table:		dw	CMD_GET_VERSION,    ProcessGetVersion
		dw	CMD_GET_DRIVE_INFO, ProcessGetDriveInfo
		dw	CMD_READ_SECTOR,    ProcessReadSector
		dw	CMD_WRITE_SECTOR,   ProcessWriteSector
		dw	CMD_OPEN_FILE,      ProcessOpenFile
		dw	CMD_GET_FILE_SIZE,  ProcessGetFileSize
		dw	CMD_READ_FILE,      ProcessReadFile
		dw	CMD_WRITE_FILE,     ProcessWriteFile
		dw	CMD_CLOSE_FILE,     ProcessCloseFile
.count:		equ	($ - .table) / 4


;******************************************************************************
;
; Function:     ProcessGetVersion
;
; Description:  Process a "get version" request.
;
;               Request must contain:
;               [+0],1: Request code
;
;               Response contains:
;               [+0],1: Request code + response flags
;               [+1],2: Version code with major and minor number
;
; Parameters:   [bp+ 4],2: Request
;               [bp+ 6],2: Size of request
;               [bp+ 8],2: Buffer to receive response
;               [bp+10],2: Size of response buffer
;
; Returns:      AX: number of bytes stored in response buffer.
;
;******************************************************************************
ProcessGetVersion:
	push	bp
	mov	bp, sp
	push	si
	push	di

	mov	si, [bp + 4]		; Set source index to request
	mov	di, [bp + 8]		; Set dest index to response
.validateRequest:
	cmp	word [bp + 6], 1	; There should be only the command
	jne	.exitFailure		;   code inside of the request
.verifyResponseSize:
	cmp	word [bp + 10], 3	; Check if there is enough space
	jb	.exitFailure		;   for inserting reponse data
.generateResult:
	mov	ah, VERSION_MAJOR	; Major version number as high byte
	mov	al, VERSION_MINOR	; Minor version number as low byte
	mov	[di + 1], ax		; Copy version number into response
.exitSuccess:
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_SUCCESS	;   set success flag and
	mov	[di], al		;   store as response code
	mov	ax, 3			; Positive response is 3 bytes
	jmp	.exit
.exitFailure:
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_FAILURE	;   set failure flag and
	mov	[di], al		;   store as response code
	mov	ax, 1			; Negative response is always 1 byte
.exit:

	pop	di
	pop	si
	pop	bp
	ret	8


;******************************************************************************
;
; Function:     ProcessGetDriveInfo
;
; Description:  Process a "get drive information" request.
;
;               Request must contain:
;               [+0],1: Request code
;               [+1],1: Drive index (0 = A:, 1 = B:, etc.)
;
;               Response contains:
;               [+0],1: Request code + response flags
;               [+1],1: Status code from BIOS call
;               [+2],1: Drive type
;               [+3],1: Number of tracks
;               [+4],1: Number of sides
;               [+5],1: Number of sectors
;
; Parameters:   [bp+ 4],2: Request
;               [bp+ 6],2: Size of request
;               [bp+ 8],2: Buffer to receive response
;               [bp+10],2: Size of response buffer
;
; Returns:      AX: number of bytes stored in response buffer.
;
; Notes:        Function 0x08 of BIOS interrupt 13h changes ES:DI. Although
;               not explicitly used in this program, ES is restored to avoid
;               strange side effects.
;
;******************************************************************************
ProcessGetDriveInfo:
	push	bp
	mov	bp, sp
	push	bx
	push	cx
	push	dx
	push	si
	push	di

	mov	si, [bp + 4]		; Set source index to request
	mov	di, [bp + 8]		; Set dest index to response
.validateRequest:
	cmp	word [bp + 6], 2	; There should be the command and
	jne	.exitFailure		;   a drive index in the request
.verifyResponseSize:
	cmp	word [bp + 10], 6	; Check if there is enough space
	jb	.exitFailure		;   for inserting reponse data
.generateResult:
	push	es			; Backup of ES, see note above
	push	di			; Backup of DI, see note above
	mov	dl, [si + 1]		; Get drive index into DL
	mov	ah, 0x08		; Function "read drive parameters"
	int	0x13			; Trigger BIOS interrupt
	pop	di			; Restore content of DI
	pop	es			; Restore content of ES
	jc	.insertInvalidResult	; Check if function call failed
.insertValidResult:
	mov	[di + 1], ah		; Get status code from AH
	mov	[di + 2], bl		; Get drive type from BL
	add	ch, 1			; CH holds last track index, add 1
	mov	[di + 3], ch		;   to make it the track count
	add	dh, 1			; DH holds last side index, add 1
	mov	[di + 4], dh		;   to make it the side count
	mov	[di + 5], cl		; Get sector count from CL
	jmp	.exitSuccess
.insertInvalidResult:
	mov	[di + 1], ah		; Get status code from AH
	mov	ax, 0xffff		; Overwrite result buffer with
	mov	[di + 2], ax		;   invalid marker to be sure it is
	mov	[di + 4], ax		;   not confused with valid data
.exitSuccess:
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_SUCCESS	;   set success flag and
	mov	[di], al		;   store as response code
	mov	ax, 6			; Positive response is 6 bytes
	jmp	.exit
.exitFailure:
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_FAILURE	;   set failure flag and
	mov	[di], al		;   store as response code
	mov	ax, 1			; Negative response is always 1 byte
.exit:

	pop	di
	pop	si
	pop	dx
	pop	cx
	pop	bx
	pop	bp
	ret	8


;******************************************************************************
;
; Function:     ProcessReadSector
;
; Description:  Process a "read sector from drive" request.
;
;               Request must contain:
;               [+0],1: Request code
;               [+1],1: Drive index (0 = A:, 1 = B:, etc.)
;               [+2],1: Track to read from
;               [+3],1: Side to read from
;               [+4],1: Sector to read
;
;               Response contains:
;               [+0],1:   Request code + response flags
;               [+1],1:   Status code from BIOS call
;               [+2],512: Data from sector
;
; Parameters:   [bp+ 4],2: Request
;               [bp+ 6],2: Size of request
;               [bp+ 8],2: Buffer to receive response
;               [bp+10],2: Size of response buffer
;
; Returns:      AX: number of bytes stored in response buffer.
;
;******************************************************************************
ProcessReadSector:
	push	bp
	mov	bp, sp
	push	bx
	push	cx
	push	dx
	push	si
	push	di

	mov	si, [bp + 4]		; Set source index to request
	mov	di, [bp + 8]		; Set dest index to response
.validateRequest:
	cmp	word [bp + 6], 5	; There should be the command and
	jne	.exitFailure		;   4 bytes of location specifier
.verifyResponseSize:
	cmp	word [bp + 10], 514	; Check if there is enough space
	jb	.exitFailure		;   for inserting reponse data
.generateResult:
	mov	dl, [si + 1]		; Place drive index in DL
	mov	ch, [si + 2]		; Place track index in CH
	mov	dh, [si + 3]		; Place side index in DH
	mov	cl, [si + 4]		; Place sector index in CL
	add	cl, 1			; Adjust for 1-based sector indices
	mov	al, 1			; Number of sectors to read in AL
	mov	bx, di			; Buffer to read data into
	add	bx, 2			;   (2 bytes offset into response)
	mov	ah, 0x02		; Function "read disk sectors"
	int	0x13			; Trigger BIOS interrupt
	mov	[di + 1], ah		; Get status code from AH
.exitSuccess:
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_SUCCESS	;   set success flag and
	mov	[di], al		;   store as response code
	mov	ax, 514			; Positive response is 514 bytes
	jmp	.exit
.exitFailure:
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_FAILURE	;   set failure flag and
	mov	[di], al		;   store as response code
	mov	ax, 1			; Negative response is always 1 byte
.exit:

	pop	di
	pop	si
	pop	dx
	pop	cx
	pop	bx
	pop	bp
	ret	8


;******************************************************************************
;
; Function:     ProcessWriteSector
;
; Description:  Process a "write sector to drive" request.
;
;               Request must contain:
;               [+0],1:   Request code
;               [+1],1:   Drive index (0 = A:, 1 = B:, etc.)
;               [+2],1:   Track to write to
;               [+3],1:   Side to write to
;               [+4],1:   Sector to write
;               [+5],512: Data of sector
;
;               Response contains:
;               [+0],1:   Request code + response flags
;               [+1],1:   Status code from BIOS call
;
; Parameters:   [bp+ 4],2: Request
;               [bp+ 6],2: Size of request
;               [bp+ 8],2: Buffer to receive response
;               [bp+10],2: Size of response buffer
;
; Returns:      AX: number of bytes stored in response buffer.
;
;******************************************************************************
ProcessWriteSector:
	push	bp
	mov	bp, sp
	push	bx
	push	cx
	push	dx
	push	si
	push	di

	mov	si, [bp + 4]		; Set source index to request
	mov	di, [bp + 8]		; Set dest index to response
.validateRequest:
	cmp	word [bp + 6], 517	; There should be the command and
	jne	.exitFailure		;   516 bytes of meta info + data
.verifyResponseSize:
	cmp	word [bp + 10], 2	; Check if there is enough space
	jb	.exitFailure		;   for inserting reponse data
.generateResult:
	mov	dl, [si + 1]		; Place drive index in DL
	mov	ch, [si + 2]		; Place track index in CH
	mov	dh, [si + 3]		; Place side index in DH
	mov	cl, [si + 4]		; Place sector index in CL
	add	cl, 1			; Adjust for 1-based sector indices
	mov	al, 1			; Number of sectors to write in AL
	mov	bx, si			; Buffer to get data from
	add	bx, 5			;   (5 bytes offset into request)
	mov	ah, 0x03		; Function "write disk sectors"
	int	0x13			; Trigger BIOS interrupt
	mov	[di + 1], ah		; Get status code from AH
.exitSuccess:
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_SUCCESS	;   set success flag and
	mov	[di], al		;   store as response code
	mov	ax, 2			; Positive response is 2 bytes
	jmp	.exit
.exitFailure:
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_FAILURE	;   set failure flag and
	mov	[di], al		;   store as response code
	mov	ax, 1			; Negative response is always 1 byte
.exit:

	pop	di
	pop	si
	pop	dx
	pop	cx
	pop	bx
	pop	bp
	ret	8


;******************************************************************************
;
; Function:     ProcessOpenFile
;
; Description:  Process an "open file" request.
;
;               Request must contain:
;               [+0],1:   Request code
;               [+1],1:   Access mode
;               [+2],260: Filename as zero-terminated string, var. length
;
;               Response contains:
;               [+0],1:   Request code + response flags
;               [+1],1:   Status code from DOS call
;               [+2],2:   File handle
;
; Parameters:   [bp+ 4],2: Request
;               [bp+ 6],2: Size of request
;               [bp+ 8],2: Buffer to receive response
;               [bp+10],2: Size of response buffer
;
; Returns:      AX: number of bytes stored in response buffer.
;
;******************************************************************************
ProcessOpenFile:
	push	bp
	mov	bp, sp
	push	bx
	push	cx
	push	dx
	push	si
	push	di

	mov	si, [bp + 4]		; Set source index to request
	mov	di, [bp + 8]		; Set dest index to response
.validateRequest:
	cmp	word [bp + 6], 4	; There should be the command and
	jb	.exitFailure		;   at least 3 more bytes
.verifyResponseSize:
	cmp	word [bp + 10], 4	; Check if there is enough space
	jb	.exitFailure		;   for inserting reponse data
.generateResult:
	mov	al, [si + 1]		; Get requested access code
	cmp	al, PRM_OPEN_FILE_READ	; Check if file should be opened
	je	.openRead		;   for reading (open existing)
	cmp	al, PRM_OPEN_FILE_WRITE	; Check if file should be opened
	je	.openWrite		;   for writing (create/truncate)
	jmp	.exitFailure		; Neither nor => invalid access mode
.openRead:
	mov	dx, si			; Get path to open
	add	dx, 2			;   (2 bytes offset into request)
	mov	al, 0			; Open access mode "read only"
	mov	ah, 0x3d		; Function "open file using handle"
	int	0x21			; Trigger DOS interrupt
	jc	.exitWithStatus		; Check if function call failed
	mov	[di + 2], ax		; Store file handle in response
	mov	al, 0			; Set success status code
	jmp	.exitWithStatus
.openWrite:
	mov	dx, si			; Get path to open
	add	dx, 2			;   (2 bytes offset into request)
	mov	cx, 0			; File attributes for regular file
	mov	ah, 0x3c		; Function "create file"
	int	0x21			; Trigger DOS interrupt
	jc	.exitWithStatus		; Check if function call failed
	mov	[di + 2], ax		; Store file handle in response
	mov	al, 0			; Set success status code
.exitWithStatus:
	mov	[di + 1], al		; Store status code in response
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_SUCCESS	;   set success flag and
	mov	[di], al		;   store as response code
	mov	ax, 4			; Positive response is always 4 bytes
	jmp	.exit
.exitFailure:
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_FAILURE	;   set failure flag and
	mov	[di], al		;   store as response code
	mov	ax, 1			; Negative response is always 1 byte
.exit:

	pop	di
	pop	si
	pop	dx
	pop	cx
	pop	bx
	pop	bp
	ret	8


;******************************************************************************
;
; Function:     ProcessGetFileSize
;
; Description:  Process a "get file size" request.
;
;               Request must contain:
;               [+0],1:   Request code
;               [+1],2:   File handle
;
;               Response contains:
;               [+0],1:   Request code + response flags
;               [+1],1:   Status code from DOS call
;               [+2],2:   Low word of file size in bytes
;               [+4],2:   High word of file size in bytes
;
; Parameters:   [bp+ 4],2: Request
;               [bp+ 6],2: Size of request
;               [bp+ 8],2: Buffer to receive response
;               [bp+10],2: Size of response buffer
;
; Returns:      AX: number of bytes stored in response buffer.
;
;******************************************************************************
ProcessGetFileSize:
	push	bp
	mov	bp, sp
	push	bx
	push	cx
	push	dx
	push	si
	push	di

	mov	si, [bp + 4]		; Set source index to request
	mov	di, [bp + 8]		; Set dest index to response
.validateRequest:
	cmp	word [bp + 6], 3	; There should be the command and
	jne	.exitFailure		;   2 bytes of meta info
.verifyResponseSize:
	cmp	word [bp + 10], 6	; Check if there is enough space
	jb	.exitFailure		;   for inserting reponse data
.generateResult:
	mov	bx, [si + 1]		; Get file handle into BX
	mov	al, 2			; Move file pointer relative to
	mov	cx, 0			;  end of file (SEEK_END) with
	mov	dx, 0			;  offset zero
	mov	ah, 0x42		; Function "move file pointer"
	int	0x21			; Trigger DOS interrupt
	jc	.exitWithStatus		; Check if function call failed
	mov	[di + 2], ax		; Store low word of file size
	mov	[di + 4], dx		; Store high word of file size
	mov	al, 0			; Move file pointer relative to
	mov	cx, 0			;  start of file (SEEK_SET) with
	mov	dx, 0			;  offset zero
	mov	ah, 0x42		; Function "move file pointer"
	int	0x21			; Trigger DOS interrupt
	jmp	.exitWithStatus
	mov	al, 0			; Set success status code
.exitWithStatus:
	mov	[di + 1], al		; Store status code in response
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_SUCCESS	;   set success flag and
	mov	[di], al		;   store as response code
	mov	ax, 6			; Positive response is always 6 bytes
	jmp	.exit
.exitFailure:
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_FAILURE	;   set failure flag and
	mov	[di], al		;   store as response code
	mov	ax, 1			; Negative response is always 1 byte
.exit:

	pop	di
	pop	si
	pop	dx
	pop	cx
	pop	bx
	pop	bp
	ret	8


;******************************************************************************
;
; Function:     ProcessReadFile
;
; Description:  Process an "read file" request.
;
;               Request must contain:
;               [+0],1:   Request code
;               [+1],2:   File handle
;
;               Response contains:
;               [+0],1:   Request code + response flags
;               [+1],1:   Status code from DOS call
;               [+2],2:   Number of bytes read
;               [+4],512: Up to 512 bytes of data
;
; Parameters:   [bp+ 4],2: Request
;               [bp+ 6],2: Size of request
;               [bp+ 8],2: Buffer to receive response
;               [bp+10],2: Size of response buffer
;
; Returns:      AX: number of bytes stored in response buffer.
;
;******************************************************************************
ProcessReadFile:
	push	bp
	mov	bp, sp
	push	bx
	push	cx
	push	dx
	push	si
	push	di

	mov	si, [bp + 4]		; Set source index to request
	mov	di, [bp + 8]		; Set dest index to response
.validateRequest:
	cmp	word [bp + 6], 3	; There should be the command and
	jne	.exitFailure		;   2 bytes of meta info
.verifyResponseSize:
	cmp	word [bp + 10], 516	; Check if there is enough space
	jb	.exitFailure		;   for inserting reponse data
.generateResult:
	mov	bx, [si + 1]		; Get file handle into BX
	mov	cx, 512			; Number of bytes to read (at max)
	mov	dx, di			; Buffer to read data to
	add	dx, 4			;   (4 bytes offset into response)
	mov	ah, 0x3f		; Function "read from file"
	int	0x21			; Trigger DOS interrupt
	jc	.exitWithStatus		; Check if function call failed
	mov	[di + 2], ax		; Store number of bytes read
	mov	al, 0			; Signal success
.exitWithStatus:
	mov	[di + 1], al		; Store status code in response
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_SUCCESS	;   set success flag and
	mov	[di], al		;   store as response code
	mov	ax, 4			; Positive response is 4 bytes
	add	ax, [di + 2]		;   plus number of bytes read
	jmp	.exit
.exitFailure:
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_FAILURE	;   set failure flag and
	mov	[di], al		;   store as response code
	mov	ax, 1			; Negative response is always 1 byte
.exit:

	pop	di
	pop	si
	pop	dx
	pop	cx
	pop	bx
	pop	bp
	ret	8


;******************************************************************************
;
; Function:     ProcessWriteFile
;
; Description:  Process a "write file" request.
;
;               Request must contain:
;               [+0],1:   Request code
;               [+1],2:   File handle
;               [+3],2:   Number of bytes to write
;               [+5],512: Up to 512 bytes of data, var. length
;
;               Response contains:
;               [+0],1:   Request code + response flags
;               [+1],1:   Status code from DOS call
;               [+2],2:   Number of bytes written
;
; Parameters:   [bp+ 4],2: Request
;               [bp+ 6],2: Size of request
;               [bp+ 8],2: Buffer to receive response
;               [bp+10],2: Size of response buffer
;
; Returns:      AX: number of bytes stored in response buffer.
;
;******************************************************************************
ProcessWriteFile:
	push	bp
	mov	bp, sp
	push	bx
	push	cx
	push	dx
	push	si
	push	di

	mov	si, [bp + 4]		; Set source index to request
	mov	di, [bp + 8]		; Set dest index to response
.validateRequest:
	cmp	word [bp + 6], 5	; There should be the command and
	jb	.exitFailure		;   at least 4 bytes of meta info
.verifyResponseSize:
	cmp	word [bp + 10], 4	; Check if there is enough space
	jb	.exitFailure		;   for inserting reponse data
.generateResult:
	mov	bx, [si + 1]		; Get file handle into BX
	mov	cx, [si + 3]		; Get number of bytes to write into CX
	mov	dx, si			; Buffer to get data from
	add	dx, 5			;   (5 bytes offset into request)
	mov	ah, 0x40		; Function "write to file"
	int	0x21			; Trigger DOS interrupt
	jc	.exitWithStatus		; Check if function call failed
	mov	[di + 2], ax		; Store number of bytes written
	mov	al, 0			; Signal success
	jmp	.exitWithStatus
.exitWithStatus:
	mov	[di + 1], al		; Store status code in response
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_SUCCESS	;   set success flag and
	mov	[di], al		;   store as response code
	mov	ax, 4			; Positive response is 4 bytes
	jmp	.exit
.exitFailure:
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_FAILURE	;   set failure flag and
	mov	[di], al		;   store as response code
	mov	ax, 1			; Negative response is always 1 byte
.exit:

	pop	di
	pop	si
	pop	dx
	pop	cx
	pop	bx
	pop	bp
	ret	8


;******************************************************************************
;
; Function:     ProcessCloseFile
;
; Description:  Process a "close file" request.
;
;               Request must contain:
;               [+0],1:   Request code
;               [+1],2:   File handle
;
;               Response contains:
;               [+0],1:   Request code + response flags
;               [+1],1:   Status code from DOS call
;
; Parameters:   [bp+ 4],2: Request
;               [bp+ 6],2: Size of request
;               [bp+ 8],2: Buffer to receive response
;               [bp+10],2: Size of response buffer
;
; Returns:      AX: number of bytes stored in response buffer.
;
;******************************************************************************
ProcessCloseFile:
	push	bp
	mov	bp, sp
	push	bx
	push	si
	push	di

	mov	si, [bp + 4]		; Set source index to request
	mov	di, [bp + 8]		; Set dest index to response
.validateRequest:
	cmp	word [bp + 6], 3	; There should be the command and
	jne	.exitFailure		;   2 bytes of meta info
.verifyResponseSize:
	cmp	word [bp + 10], 2	; Check if there is enough space
	jb	.exitFailure		;   for inserting reponse data
.generateResult:
	mov	bx, [si + 1]		; Get file handle into BX
	mov	ah, 0x3e		; Function "close file using handle"
	int	0x21			; Trigger DOS interrupt
	jc	.exitWithStatus		; Check if function call failed
	mov	al, 0			; Signal success
.exitWithStatus:
	mov	[di + 1], al		; Store status code in response
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_SUCCESS	;   set success flag and
	mov	[di], al		;   store as response code
	mov	ax, 2			; Positive response is 2 bytes
	jmp	.exit
.exitFailure:
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_FAILURE	;   set failure flag and
	mov	[di], al		;   store as response code
	mov	ax, 1			; Negative response is always 1 byte
.exit:

	pop	di
	pop	si
	pop	bx
	pop	bp
	ret	8


;******************************************************************************
;
; Function:     ProcessInvalidCommand
;
; Description:  Process a request with unknown command.
;
; Parameters:   [bp+ 4],2: Request
;               [bp+ 6],2: Size of request
;               [bp+ 8],2: Buffer to receive response
;               [bp+10],2: Size of response buffer
;
; Returns:      AX: number of bytes stored in response buffer.
;
;******************************************************************************
ProcessInvalidCommand:
	push	bp
	mov	bp, sp
	push	si
	push	di

	mov	si, [bp + 4]		; Set source index to request
	mov	di, [bp + 8]		; Set dest index to response
.exitFailure:
	mov	al, [si]		; Get request code,
	or	al, RES_FLAG_FAILURE	;   set failure flag and
	mov	[di], al		;   store as response code
	mov	ax, 1			; Negative response is always 1 byte
.exit:

	pop	di
	pop	si
	pop	bp
	ret	8


;******************************************************************************
;
; Function:     WriteResponse
;
; Description:  Writes a response to the specified COM port.
;
;               Writes 2 bytes to form the number of bytes in packet.
;               Then writes that number of bytes.
;               When writing fails, the call returns.
;
; Parameters:   [bp+4],2: COM port to send on
;               [bp+6],2: Response
;               [bp+8],2: Size of response
;
; Returns:      Void.
;
;******************************************************************************
WriteResponse:
	push	bp
	mov	bp, sp
	pushf
	push	ax
	push	cx
	push	dx

	mov	dx, [bp + 4]		; COM port to write to
	mov	cx, [bp + 8]		; Number of bytes to send
.writeSizeLow:
	mov	al, cl			; Send low byte first
	mov	ah, 0x01		; Function "send character"
	int	0x14			; Trigger BIOS interrupt
	test	ah, 0x80		; Test for failure flag
	jnz	.exit			; If failed, then stop
.writeSizeHigh:
	mov	al, ch			; Send high byte second
	mov	ah, 0x01		; Function "send character"
	int	0x14			; Trigger BIOS interrupt
	test	ah, 0x80		; Test for failure flag
	jnz	.exit			; If failed, then stop
.writeData:
	cmp	cx, 0			; Check if it is a null write,
	je	.exit			;   then skip it
	cld				; Increase SI when reading
	mov	si, [bp + 6]		; Get pointer to buffer
.send:
	lodsb				; Read next byte to AL
	mov	ah, 0x01		; Function "send character"
	int	0x14			; Trigger BIOS interrupt
	test	ah, 0x80		; Test for failure flag
	jnz	.exit			; If failed, then stop
	loop	.send			; Loop for next byte
.exit:
	pop	dx
	pop	cx
	pop	ax
	popf
	pop	bp
	ret	6


;******************************************************************************
; String table
;******************************************************************************
banner_1:	db	'Toolbox version $'
banner_2:	db	' started.', 0x0d, 0x0a, '$'
