#!/usr/bin/perl

# memcheck.pl -- checks if calls to MALLOC match calls to FREE.
#
# Usage: memcheck.pl <text file with program debug output>
#
# Finds the following errors:
# - allocated but never freed block (aka. memory leak)
# - free of block that was already freed (aka. double free)
# - free of block that was never allocated (aka. stupid)
# Prints statistics.

use strict;

if (scalar @ARGV != 1)
{
    print "Usage: memcheck.pl <text file with program debug output>\n";
    exit 1;
}

open(my $fh, "< $ARGV[0]") or die "Could not open input file";
my @lines = <$fh>;
close($fh);

my %allocations;
my %places;
my $totalAllocs;
my $totalBytes;
my $defects;
foreach (@lines)
{
    if (m/^MALLOC: (\d+) bytes at (0x[0-9a-fA-F]+) from (.*[\\\/])?([^\r\n]+)/)
    {
        $allocations{$2} = $1;
        $places{$2} = $4;
        $totalAllocs++;
        $totalBytes += $1;
    }
    elsif (m/^FREE: (0x[0-9a-fA-F]+) from (.*[\\\/])?([^\r\n]+)/)
    {
        if (not exists $allocations{$1})
        {
            print "ERROR: free of never allocated block at $1 from $3\n";
        }
        elsif ($allocations{$1} == 0)
        {
            print "ERROR: double free at $1 from $3\n";
            $defects++;
        }
        $allocations{$1} = 0;
    }
}

foreach (keys %allocations)
{
    my $bytes = $allocations{$_};
    if ($bytes != 0)
    {
        print "ERROR: memory leak at $_ of $bytes bytes from $places{$_}.\n";
        $defects++;
    }
}

print "INFO: allocated $totalBytes bytes in total using $totalAllocs allocations.\n";
print "INFO: no defects found.\n" if ($defects == 0);
