#include <stdio.h>
#include <windows.h>

#define CHAR_BURST_SIZE   64
#define CHAR_BURST_START  'A'
#define CHAR_BURST_END    'Z'
#define CHAR_VALUE        'A'

struct MenuItem {
	const char *lpTitle;
	int iKey;
	void(*lpHandler)(HANDLE hDevice);
};

static void SetRts(HANDLE hDevice);
static void ClearRts(HANDLE hDevice);
static void SetDtr(HANDLE hDevice);
static void ClearDtr(HANDLE hDevice);
static void SendSingleChar(HANDLE hDevice);
static void SendCharBurst(HANDLE hDevice);
static void Quit(HANDLE hDevice);

static const struct MenuItem Menu[] =
{
	/* Title                    Key   Handler        */
	{ "Set RTS",                'R',  SetRts         },
	{ "Clear RTS",              'r',  ClearRts       },
	{ "Set DTR",                'D',  SetDtr         },
	{ "Clear DTR",              'd',  ClearDtr       },
	{ "Send single character",  's',  SendSingleChar },
	{ "Send character burst",   'b',  SendCharBurst  },
	{ "Quit",                   'q',  Quit           },
};

static void SetRts(HANDLE hDevice)
{
	if (EscapeCommFunction(hDevice, SETRTS))
	{
		printf("RTS set.\n");
	}
	else
	{
		fprintf(stderr, "EscapeCommFunction failed: %lu\n", GetLastError());
	}
}

static void ClearRts(HANDLE hDevice)
{
	if (EscapeCommFunction(hDevice, CLRRTS))
	{
		printf("RTS cleared.\n");
	}
	else
	{
		fprintf(stderr, "EscapeCommFunction failed: %lu\n", GetLastError());
	}
}

static void SetDtr(HANDLE hDevice)
{
	if (EscapeCommFunction(hDevice, SETDTR))
	{
		printf("DTR set.\n");
	}
	else
	{
		fprintf(stderr, "EscapeCommFunction failed: %lu\n", GetLastError());
	}
}

static void ClearDtr(HANDLE hDevice)
{
	if (EscapeCommFunction(hDevice, CLRDTR))
	{
		printf("DTR cleared.\n");
	}
	else
	{
		fprintf(stderr, "EscapeCommFunction failed: %lu\n", GetLastError());
	}
}

static void SendSingleChar(HANDLE hDevice)
{
	char ch = CHAR_VALUE;
	DWORD dwBytes;
	if (WriteFile(hDevice, &ch, 1, &dwBytes, NULL))
	{
		if (dwBytes == 1)
		{
			printf("Character sent.\n");
		}
		else
		{
			printf("Short write (%i/%i).\n", dwBytes, 1);
		}
	}
	else
	{
		fprintf(stderr, "WriteFile failed: %lu\n", GetLastError());
	}
}

static void SendCharBurst(HANDLE hDevice)
{
	char buf[CHAR_BURST_SIZE];

	int ch = CHAR_BURST_START;
	for (int i = 0; i < CHAR_BURST_SIZE; i++)
	{
		buf[i] = ch;
		ch = (ch + 1 <= CHAR_BURST_END) ? (ch + 1) : CHAR_BURST_START;
	}

	DWORD dwBytes;
	if (WriteFile(hDevice, buf, CHAR_BURST_SIZE, &dwBytes, NULL))
	{
		if (dwBytes == CHAR_BURST_SIZE)
		{
			printf("%i characters sent.\n", CHAR_BURST_SIZE);
		}
		else
		{
			printf("Short write (%i/%i).\n", dwBytes, CHAR_BURST_SIZE);
		}
	}
	else
	{
		fprintf(stderr, "WriteFile failed: %lu\n", GetLastError());
	}
}

static BOOL fRunning;

static void Quit(HANDLE hDevice)
{
	fRunning = FALSE;
}

static HANDLE OpenCommDevice(LPCSTR lpPortName)
{
	BOOL fFailed = FALSE;

	HANDLE hResult = CreateFile(lpPortName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
	fFailed = hResult == INVALID_HANDLE_VALUE;

	if (hResult != INVALID_HANDLE_VALUE)
	{
		DCB dcb;
		ZeroMemory(&dcb, sizeof(dcb));
		dcb.DCBlength = sizeof(dcb);
		dcb.BaudRate = CBR_9600;
		dcb.ByteSize = 8;
		dcb.fParity = NOPARITY;
		dcb.StopBits = ONESTOPBIT;
		dcb.fBinary = TRUE;
		
		fFailed = !SetCommState(hResult, &dcb);
	}

	if (fFailed && hResult != INVALID_HANDLE_VALUE)
	{
		CloseHandle(hResult);
		hResult = INVALID_HANDLE_VALUE;
	}

	return hResult;
}

int main(void)
{
	HANDLE hDevice = OpenCommDevice("COM4");
	if (hDevice != INVALID_HANDLE_VALUE)
	{
		fRunning = TRUE;
		while (fRunning)
		{
			for (int i = 0; i < ARRAYSIZE(Menu); i++)
			{
				printf("%c) %-22s\n", Menu[i].iKey, Menu[i].lpTitle);
			}
			printf("=> ");
			fflush(stdout);

			char buf[100];
			fgets(buf, sizeof(buf), stdin);
			for (int i = 0; i < ARRAYSIZE(Menu); i++)
			{
				if (buf[0] == Menu[i].iKey)
				{
					Menu[i].lpHandler(hDevice);
					printf("\n");
				}
			}
		}

		CloseHandle(hDevice);
	}

	return 0;
}
