/******************************************************************************
*
* Filename:     serial-send.c
*
* Description:  Serial sender
*               Reads binary data from a file and sends it to serial port.
*               This program uses DTR flow control, because the receiver side
*               seems to require this when using INT 14 functions.
*               Timing can be fine-tuned using buffer size and delay values.
*
******************************************************************************/

#include <stdio.h>
#include <windows.h>

// COM device configuration
#define CFG_BAUDRATE     9600
#define CFG_DATABITS     8
#define CFG_PARITY       NOPARITY
#define CFG_STOPBITS     ONESTOPBIT
// Data transfer configuration
#define CFG_BUFFER_SIZE  16   // bytes
#define CFG_DELAY        100  // milliseconds

struct buffer {
    BYTE data[CFG_BUFFER_SIZE];
    DWORD length;
};

const char *get_last_error(void)
{
    static char buffer[100];

    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,  // dwFlags
                  NULL,                        // lpSource
                  GetLastError(),              // dwMessageId
                  0,                           // dwLanguageId
                  buffer,                      // lpBuffer
                  sizeof(buffer),              // nSize
                  NULL);                       // Arguments

    return buffer;
}

void init_dcb(DCB *dcb)
{
    ZeroMemory(dcb, sizeof(DCB));
    dcb->DCBlength   = sizeof(DCB);
    dcb->BaudRate    = CFG_BAUDRATE;
    dcb->fBinary     = TRUE;
    dcb->ByteSize    = CFG_DATABITS;
    dcb->Parity      = CFG_PARITY;
    dcb->StopBits    = CFG_STOPBITS;
    dcb->fDtrControl = DTR_CONTROL_ENABLE;
}

HANDLE open_port(const char *name)
{
    return CreateFile(name,           // lpFileName
                      GENERIC_WRITE,  // dwDesiredAccess
                      0,              // dwShareMode
                      NULL,           // lpSecurityAttributes
                      OPEN_EXISTING,  // dwCreationDisposition
                      0,              // dwFlagsAndAttributes
                      NULL);          // hTemplateFile
}

HANDLE open_file(const char *name)
{
    return CreateFile(name,           // lpFileName
                      GENERIC_READ,   // dwDesiredAccess
                      0,              // dwShareMode
                      NULL,           // lpSecurityAttributes
                      OPEN_EXISTING,  // dwCreationDisposition
                      0,              // dwFlagsAndAttributes
                      NULL);          // hTemplateFile
}

BOOL read_block(HANDLE handle, struct buffer *block)
{
    return ReadFile(handle,               // hFile,
                    block->data,          // lpBuffer
                    sizeof(block->data),  // nNumberOfBytesToRead
                    &block->length,       // lpNumberOfBytesRead
                    NULL);                // lpOverlapped
}

BOOL write_block(HANDLE handle, const struct buffer *block)
{
    BOOL failed = FALSE;
    LPCBYTE pos = block->data;
    DWORD outstanding = block->length;

    while (!failed && outstanding > 0) {
        DWORD written;

        failed = !WriteFile(handle,       // hFile
                            pos,          // lpBuffer
                            outstanding,  // nNumberOfBytesToWrite
                            &written,     // lpNumberOfBytesWritten
                            NULL);        // lpOverlapped
        pos += written;
        outstanding -= written;
    }

    return !failed;
}

BOOL send_data(HANDLE in, HANDLE out)
{
    BOOL failed = FALSE;
    BOOL done = FALSE;

    // loop until either done or failed
    while (!done && !failed) {
        struct buffer block;

        // read next block of data
        if (!read_block(in, &block)) {
            fprintf(stderr, "Read failed: %s\n", get_last_error());
            failed = TRUE;
            break;
        }
        // check if end of input file is reached
        if (block.length == 0) {
            done = TRUE;
            break;
        }
        // print status and write block
        printf(" [%lu", block.length);
        if (!write_block(out, &block)) {
            fprintf(stderr, "Write failed: %s\n", get_last_error());
            failed = TRUE;
            break;
        }
        printf("]");
        // if configured: sleep for specified time
        if (CFG_DELAY > 0) {
            Sleep(CFG_DELAY);
        }
    }

    return !failed && done;
}

int main(int argc, char *argv[])
{
    BOOL failed = FALSE;

    // validate parameters
    if (argc != 3) {
        fprintf(stderr, "usage: %s <file> <port>\n", argv[0]);
        failed = TRUE;
    }

    // open binary file for reading
    const char *filename = argv[1];
    HANDLE file = INVALID_HANDLE_VALUE;
    if (!failed) {
        file = open_file(filename);
        if (file == INVALID_HANDLE_VALUE) {
            fprintf(stderr, "Opening input file \"%s\" failed: %s\n",
                filename, get_last_error());
            failed = TRUE;
        }
    }

    // open COM port for sending
    const char *devname = argv[2];
    HANDLE dev = INVALID_HANDLE_VALUE;
    if (!failed) {
        dev = open_port(devname);
        if (dev == INVALID_HANDLE_VALUE) {
            fprintf(stderr, "Opening COM device \"%s\" failed: %s\n",
                devname, get_last_error());
            failed = TRUE;
        }
    }

    // configure COM port
    if (!failed) {
        DCB dcb;
        init_dcb(&dcb);
        if (!SetCommState(dev, &dcb)) {
            fprintf(stderr, "Configuring COM device \"%s\" failed: %s\n",
                devname, get_last_error());
            failed = TRUE;
        }
    }

    // send data
    if (!failed) {
        printf("Sending data:");
        if (send_data(file, dev)) {
            printf(" done.\n");
        } else {
            printf(" failed!\n");
            failed = TRUE;
        }
    }

    // cleanup as necessary
    if (dev != INVALID_HANDLE_VALUE) {
        CloseHandle(dev);
    }
    if (file != INVALID_HANDLE_VALUE) {
        CloseHandle(file);
    }

    return failed ? EXIT_FAILURE : EXIT_SUCCESS;
}
